"""
Modified from MinDScAIP: Minimally biased Differentiable Scaled Attention Interatomic Potential
Credit: Ryan Liu
"""

from __future__ import annotations

from typing import TYPE_CHECKING

import torch

if TYPE_CHECKING:
    from fairchem.core.datasets.atomic_data import AtomicData


def safe_norm(
    x: torch.Tensor,
    dim: int = -1,
    keepdim: bool = False,
    eps: float = 1e-12,
) -> torch.Tensor:
    """
    Computes the norm of a tensor with a small epsilon to avoid division by zero.
    Args:
        x: The input tensor.
        dim: The dimension to reduce.
        keepdim: Whether to keep the reduced dimension.
        eps: The epsilon value.
    Returns:
        The norm of the input tensor.
    """
    vec_norm_sq = x.square().sum(dim=dim, keepdim=keepdim)
    vec_norm = vec_norm_sq.clamp_min(eps).sqrt()
    vec_norm = torch.where(vec_norm_sq <= eps, torch.zeros_like(vec_norm), vec_norm)
    return vec_norm


def safe_normalize(
    x: torch.Tensor,
    dim: int = -1,
    eps: float = 1e-12,
) -> torch.Tensor:
    """
    Computes the normalized vector with a small epsilon to avoid division by zero.
    Args:
        x: The input tensor.
        dim: The dimension to reduce.
        eps: The epsilon value.
    Returns:
        The L2 normalized tensor.
    """
    vec_norm_sq = x.square().sum(dim=dim, keepdim=True)
    vec_norm = vec_norm_sq.clamp_min(eps).sqrt()
    norm_vec = torch.where(vec_norm_sq <= eps, torch.zeros_like(x), x / vec_norm)
    return norm_vec


def envelope_fn(
    x: torch.Tensor,
    envelope: bool = True,
) -> torch.Tensor:
    """
    Computes the envelope function in log space that smoothly vanishes to -inf at x = 1.
    Args:
        x: The input tensor.
        envelope: Whether to use the envelope function. Default: True
    Returns:
        The envelope function in log space.
    """
    if envelope:
        env = -x.pow(2) / (1 - x.pow(2))
    else:
        env = torch.zeros_like(x)
    return torch.where(x < 1, env, -torch.inf)


def shifted_sine(x: torch.Tensor) -> torch.Tensor:
    """
    Shifted sine function for the low memory soft knn. Designed such that the behavior
    matches sigmoid for small x and the step function for large x.
    Args:
        x: the input tensor
    Returns:
        y: the shifted sine function value
    """
    return (
        0.5 * torch.where(x.abs() < torch.pi, torch.sin(0.5 * x), torch.sign(x)) + 0.5
    )


def soft_rank(
    dist: torch.Tensor,
    scale: float,
) -> torch.Tensor:
    """
    calculate the soft rankings for the soft knn
    Args:
        dist: the pairwise distance tensor
        scale: the scale factor for the sigmoid function (Å).
    Returns:
        ranks: the soft rankings
    """
    ranks = torch.sigmoid((dist[:, :, None] - dist[:, None, :]) / scale).sum(dim=-1)
    return ranks


def hard_rank(
    dist: torch.Tensor,
) -> torch.Tensor:
    """
    calculate the hard rankings for the hard knn
    Args:
        dist: the pairwise distance tensor
    Returns:
        ranks: the hard rankings
    """
    ranks = torch.empty_like(dist)
    ranks[
        torch.arange(dist.size(0), device=dist.device)[:, None],
        torch.argsort(dist, dim=-1),
    ] = torch.arange(dist.size(-1), device=dist.device, dtype=dist.dtype)
    return ranks


def soft_rank_low_mem(
    dist: torch.Tensor,
    k: int,
    scale: float,
    delta: int = 20,
) -> torch.Tensor:
    """
    calculate the soft rankings for the soft knn. Approximate with low memory by
    truncating the distance matrix to be [0, k + delta]. This is not exact but is a good
    approximation. It is valid when the difference of distance at k+delta and k is
    larger than pi * scale.
    Args:
        dist: the pairwise distance tensor
        k: the number of neighbors
        scale: the scale factor for the shifted sine function (Å).
        delta: the delta factor for the truncation
    Returns:
        ranks: the soft rankings
    """
    sorted_dist, indicies = torch.sort(dist, dim=-1)
    ranks_T = shifted_sine(
        (sorted_dist[:, : k + delta, None] - sorted_dist[:, None, : k + delta]) / scale
    ).sum(dim=-1)
    ranks = torch.full_like(dist, torch.inf)
    ranks[
        torch.arange(dist.size(0), device=dist.device)[:, None],
        indicies[:, : k + delta],
    ] = ranks_T
    return ranks


@torch.jit.script
def build_radius_graph(
    pos: torch.Tensor,
    cell: torch.Tensor,
    image_id: torch.Tensor,
    cutoff: float,
    start_index: int,
    device: torch.device,
    k: int = 30,
    soft: bool = False,
    sigmoid_scale: float = 0.2,
    lse_scale: float = 0.1,
    use_low_mem: bool = False,
) -> tuple[
    torch.Tensor,
    torch.Tensor,
    torch.Tensor,
    torch.Tensor,
    torch.Tensor,
    torch.Tensor,
]:
    """
    construct the biknn radius graph for one system.
    Args:
        pos: the atomic positions tensor
        cell: the cell tensor for the periodic boundary condition
        image_id: the image identifier for different PBC images
        cutoff: the cutoff distance in Angstrom
        start_index: the starting index of the system in the batch
        device: the device on which the tensors are allocated
        k: the number of neighbors
        soft: the flag for the soft knn
        sigmoid_scale: the scale factor for the sigmoid function
        lse_scale: the scale factor for the log-sum-exp function
        use_low_mem: the flag for the low memory soft knn
    Returns:
        index1: the source index of the neighbors
        index2: the destination index of the neighbors
        index1_rank: the rank of the edge in source neighbors by envelope function
        index2_rank: the rank of the edge in destination neighbors by envelope function
        disp: the displacement vector of the neighbors
        env: the envelope vector of the neighbors
    """
    N = pos.size(0)
    M = image_id.size(0)
    # calculate the displacements while taking into account the PBC
    src_pos = pos[:, None] + torch.mm(image_id, cell)[None, :]
    disp = src_pos[None, :, :, :] - pos[:, None, None, :]
    dist = safe_norm(disp, dim=-1)
    dist_T = dist.transpose(0, 1).contiguous()
    # compute the rankings, depending on the soft or hard knn
    if soft:
        # calculate the rankings in a soft manner
        if use_low_mem:
            # use low memory soft knn
            src_ranks = soft_rank_low_mem(dist.view(N, N * M), k, sigmoid_scale).view(
                N, N, M
            )
            dst_ranks = (
                soft_rank_low_mem(dist_T.view(N, N * M), k, sigmoid_scale)
                .view(N, N, M)
                .transpose(0, 1)
            )
        else:
            # use full soft knn
            src_ranks = soft_rank(dist.view(N, N * M), sigmoid_scale).view(N, N, M)
            dst_ranks = (
                (soft_rank(dist_T.view(N, N * M), sigmoid_scale))
                .view(N, N, M)
                .transpose(0, 1)
            )
        # env is the soft maximum of the source and destination rankings and the
        # distance normalized by the radius cutoff.
        env = torch.stack([src_ranks / k, dst_ranks / k, dist / cutoff], dim=0)
        env = lse_scale * torch.logsumexp(env / lse_scale, dim=0)
    else:
        # calculate the rankings in a hard manner
        src_ranks = hard_rank((dist).view(N, N * M)).view(N, N, M)
        dst_ranks = hard_rank((dist_T).view(N, N * M)).view(N, N, M).transpose(0, 1)
        # env is the hard maximum of the source and destination rankings and the
        # distance normalized by the radius cutoff.
        env = torch.stack([src_ranks / k, dst_ranks / k, dist / cutoff], dim=0)
        env = torch.amax(env, dim=0)
    # set the envelope to zero for self-loops
    env.masked_fill_(dist == 0.0, 0.0)
    # sort the distances of source and destintion neighbors
    index = torch.arange(N, device=device)[:, None]
    ranks = torch.arange(M * N, device=device)[None, :]
    # ranks are the ranks of the atoms within the neighbors, i.e. the j-th source
    # neighbor of i-th atom should have index1 of i and index1_rank of j.
    index1_rank = torch.empty((N, N, M), device=device, dtype=torch.long)
    src_argsort = torch.argsort(env.view(N, N * M), dim=1)
    index1_rank[index, src_argsort // M, src_argsort % M] = ranks
    index2_rank = torch.empty((N, N, M), device=device, dtype=torch.long)
    dst_argsort = torch.argsort(env.transpose(0, 1).reshape(N, N * M), dim=1)
    index2_rank[dst_argsort // M, index, dst_argsort % M] = ranks
    # compute the mask of the neighbors
    mask = env < 1.0
    # select the neighbors within the cutoff
    index1, index2, index3 = torch.where(mask)
    index1_rank = index1_rank[index1, index2, index3]
    index2_rank = index2_rank[index1, index2, index3]
    disp = disp[index1, index2, index3]
    env = env[index1, index2, index3]
    # add the start index
    index1 = index1 + start_index
    index2 = index2 + start_index
    return (
        index1,
        index2,
        index1_rank,
        index2_rank,
        disp,
        env,
    )


def batched_radius_graph(
    pos_list: list[torch.Tensor],
    cell_list: list[torch.Tensor],
    image_id_list: list[torch.Tensor],
    N: int,
    natoms: torch.Tensor,
    knn_k: int,
    knn_soft: bool,
    knn_sigmoid_scale: float,
    knn_lse_scale: float,
    knn_use_low_mem: bool,
    knn_pad_size: int,
    cutoff: float,
    device: torch.device,
) -> tuple[
    torch.Tensor,
    torch.Tensor,
    torch.Tensor,
    torch.Tensor,
    torch.Tensor,
    torch.Tensor,
]:
    """
    calculate the biknn radius graph for the batch of systems
    Args:
        pos_list: the list of atomic positions tensors
        anum_list: the list of atomic number tensors
        cell_list: the list of cell tensors
        image_id_list: the list of image identifier tensors
        N: the total number of atoms in the batch
        natoms: the number of atoms in each system
        knn_params: the parameters for the knn algorithm
        cutoff: the cutoff distance in Angstrom
        device: the device on which the tensors are allocated
    Returns:
        padded_disp: the padded displacement tensor
        src_env: the source envelope tensor
        dst_env: the destination envelope tensor
        src_index: the destination layout to source layout index tensor
        dst_index: the source layout to destination layout index tensor
        edge_index: the edge index tensor
    """
    # calculate the starting index of each system
    start_idxs = torch.cumsum(natoms, dim=0) - natoms
    # build the biknn radius graph for each system
    index1, index2, index1_rank, index2_rank, disp, env = map(
        torch.cat,
        zip(
            *[
                build_radius_graph(
                    pos,
                    cell,
                    image_id,
                    cutoff,
                    start_idx,
                    device,
                    knn_k,
                    knn_soft,
                    knn_sigmoid_scale,
                    knn_lse_scale,
                    knn_use_low_mem,
                )
                for pos, cell, image_id, start_idx in zip(
                    pos_list, cell_list, image_id_list, start_idxs
                )
            ]
        ),
    )
    # if soft knn, pad the tensors with the maximum number of neighbors.
    padsize = knn_pad_size
    # initialize the padded tensors
    padded_disp = torch.zeros((N, padsize, 3), device=device)
    src_env = torch.full((N, padsize), torch.inf, device=device)
    dst_env = torch.full((N, padsize), torch.inf, device=device)
    edge_index = torch.zeros((2, N, padsize), device=device, dtype=torch.long)
    src_index = torch.zeros((2, N, padsize), device=device, dtype=torch.long)
    dst_index = torch.zeros((2, N, padsize), device=device, dtype=torch.long)
    # fill the padded tensors
    padded_disp[index1, index1_rank] = disp
    src_env[index1, index1_rank] = env
    dst_env[index2, index2_rank] = env
    edge_index[0, index1, index1_rank] = index1
    edge_index[1, index1, index1_rank] = index2
    # the flipping index for switching between in source neighbors layout and
    # destination neighbors layout. Since index1 is the source atom's index and index2
    # is the destination atom's index, the edge that were placed at
    # [index1, index1_rank] in source neighbors layout should be placed at
    # [index2, index2_rank]
    src_index[0, index1, index1_rank] = index2
    src_index[1, index1, index1_rank] = index2_rank
    dst_index[0, index2, index2_rank] = index1
    dst_index[1, index2, index2_rank] = index1_rank

    return (
        padded_disp,
        src_env,
        dst_env,
        src_index,
        dst_index,
        edge_index,
    )


@torch.compiler.disable(recursive=True)
def biknn_radius_graph(
    data: AtomicData,
    cutoff: float,
    knn_k: int,
    knn_soft: bool,
    knn_sigmoid_scale: float,
    knn_lse_scale: float,
    knn_use_low_mem: bool,
    knn_pad_size: int,
    use_pbc: bool,
    device: torch.device,
) -> tuple[
    torch.Tensor,
    torch.Tensor,
    torch.Tensor,
    torch.Tensor,
    torch.Tensor,
    torch.Tensor,
]:
    """
    function to construct the biknn radius graph for the batch of systems. This function
    calculates the number of images to be included in the PBC and constructs the
    image identifier list and call the batched_radius_graph function to perform the
    construction.
    Args:
        data: the `torch_geometric.data.Data` object containing the atomic information
        cutoff: the cutoff distance in Angstrom
        knn_params: the parameters for the knn algorithm
        use_pbc: the flag for the periodic boundary condition
        device: the device on which the tensors are allocated
    Returns:
        padded_disp: the padded displacement tensor
        src_env: the source envelope tensor
        dst_env: the destination envelope tensor
        src_index: the destination layout to source layout index tensor
        dst_index: the source layout to destination layout index tensor
        edge_index: the edge index tensor
    """
    # if PBC is used, construct the image identifier list by including all images within
    # the cutoff distance. Adopted from FairChem repository
    if use_pbc:
        cross_a2a3 = torch.cross(data.cell[:, 1], data.cell[:, 2], dim=-1)
        cell_vol = torch.sum(data.cell[:, 0] * cross_a2a3, dim=-1, keepdim=True)

        inv_min_dist_a1 = safe_norm(cross_a2a3 / cell_vol, dim=-1)
        rep_a1 = torch.ceil(cutoff * inv_min_dist_a1)

        cross_a3a1 = torch.cross(data.cell[:, 2], data.cell[:, 0], dim=-1)
        inv_min_dist_a2 = safe_norm(cross_a3a1 / cell_vol, dim=-1)
        rep_a2 = torch.ceil(cutoff * inv_min_dist_a2)

        cross_a1a2 = torch.cross(data.cell[:, 0], data.cell[:, 1], dim=-1)
        inv_min_dist_a3 = safe_norm(cross_a1a2 / cell_vol, dim=-1)
        rep_a3 = torch.ceil(cutoff * inv_min_dist_a3)
        rep_a1, rep_a2, rep_a3 = rep_a1.tolist(), rep_a2.tolist(), rep_a3.tolist()
    else:
        rep_a1 = rep_a2 = rep_a3 = [0] * data.natoms.size(0)
    image_id_list: list[torch.Tensor] = [
        torch.cartesian_prod(
            *[
                torch.arange(
                    -rep,
                    rep + 1,
                    device=device,
                    dtype=torch.get_default_dtype(),
                )
                for rep in reps
            ]
        )
        for reps in zip(rep_a1, rep_a2, rep_a3)
    ]
    pos_list: list[torch.Tensor] = list(torch.split(data.pos, data.natoms.tolist()))
    cell_list: list[torch.Tensor] = list(data.cell)

    # call to the batched_radius_graph function to perform per-system biknn radius
    # graph construction.
    return batched_radius_graph(
        pos_list,
        cell_list,
        image_id_list,
        data.pos.size(0),
        data.natoms,
        knn_k,
        knn_soft,
        knn_sigmoid_scale,
        knn_lse_scale,
        knn_use_low_mem,
        knn_pad_size,
        cutoff,
        device,
    )



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

import os
from typing import TYPE_CHECKING, Any, ClassVar

import numpy as np
import pandas as pd
from tqdm import tqdm

from fairchem.core.components.calculate import CalculateRunner

if TYPE_CHECKING:
    from ase.calculators.calculator import Calculator

    from fairchem.core.datasets import AseDBDataset


class AdsorptionSinglePointRunner(CalculateRunner):
    """
    Singlepoint evaluator for OC20 Adsorption systems

    OC20 originally reported adsorption energies. This runner provides the
    ability to compute adsorption energy S2EF numbers by referencing to the
    provided slab atoms object. Total energy S2EF evaluations are also
    possible.
    """

    result_glob_pattern: ClassVar[str] = "adsorption-singlepoint_*-*.json.gz"

    def __init__(
        self,
        calculator: Calculator,
        input_data: AseDBDataset,
        evaluate_total_energy: bool = False,
        adsorption_energy_model: bool = False,
    ):
        """
        Initialize the AdsorptionSinglePointRunner

        Args:
            calculator: ASE calculator to use for energy and force calculations
            input_data: Dataset containing atomic structures to process
            evaluate_total_energy: Whether to evaluate total energies
            adsorption_energy_model: Whether the provided calculator is an adsorption energy model
        """
        self.evaluate_total_energy = evaluate_total_energy
        self.adsorption_energy_model = adsorption_energy_model
        if self.adsorption_energy_model:
            assert (
                not self.evaluate_total_energy
            ), "Total energy evals not available for adsorption energy models"
        super().__init__(calculator=calculator, input_data=input_data)

    def calculate(self, job_num: int = 0, num_jobs: int = 1) -> list[dict[str, Any]]:
        """
        Args:
            job_num (int, optional): Current job number in array job. Defaults to 0.
            num_jobs (int, optional): Total number of jobs in array. Defaults to 1.

        Returns:
            list[dict[str, Any]] - List of dictionaries containing calculation results
        """
        all_results = []
        chunk_indices = np.array_split(range(len(self.input_data)), num_jobs)[job_num]
        for i in tqdm(chunk_indices, desc="Running singlepoints"):
            atoms = self.input_data.get_atoms(i)
            identifier = atoms.info["identifier"]
            # extract targets
            dft_energy = atoms.get_potential_energy()
            dft_forces = atoms.get_forces()
            dft_slab_atoms = atoms.info["dft_slab_atoms"]
            dft_slab_energy = atoms.info["dft_slab_energy"]
            gas_reference_energy = atoms.info["gas_ref"]
            constraints = atoms.constraints
            if len(constraints) != 0:
                fixed_mask = atoms.constraints[0].index
            else:
                fixed_mask = []
            free_mask = np.array([i for i in range(len(atoms)) if i not in fixed_mask])
            dft_adsorption_energy = dft_energy - dft_slab_energy - gas_reference_energy

            # compute ml predictions
            atoms.calc = self.calculator
            pred_energy = atoms.get_potential_energy()
            pred_forces = atoms.get_forces()

            results = {
                "sid": identifier,
                "forces": pred_forces[free_mask],
                "forces_target": dft_forces[free_mask],
            }
            if self.adsorption_energy_model:
                results["energy_target"] = dft_adsorption_energy
                results["energy"] = pred_energy
            else:
                if self.evaluate_total_energy:
                    results["energy_target"] = dft_energy
                    results["energy"] = pred_energy
                else:
                    dft_slab_atoms.calc = self.calculator
                    results["energy_target"] = dft_adsorption_energy
                    results["energy"] = (
                        pred_energy
                        - dft_slab_atoms.get_potential_energy()
                        - gas_reference_energy
                    )
            all_results.append(results)

        return all_results

    def write_results(
        self,
        results: list[dict[str, Any]],
        results_dir: str,
        job_num: int = 0,
        num_jobs: int = 1,
    ) -> None:
        """Write calculation results to a compressed JSON file.

        Args:
            results: List of dictionaries containing elastic properties
            results_dir: Directory path where results will be saved
            job_num: Index of the current job
            num_jobs: Total number of jobs
        """
        results_df = pd.DataFrame(results)
        results_df.to_json(
            os.path.join(
                results_dir, f"adsorption-singlepoint_{num_jobs}-{job_num}.json.gz"
            )
        )

    def save_state(self, checkpoint_location: str, is_preemption: bool = False) -> bool:
        return True



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

import logging
import warnings
from collections import defaultdict
from pathlib import Path
from typing import TYPE_CHECKING, Any

import numpy as np
import torch
from torch import nn
from torch.utils.data import DataLoader, Dataset
from tqdm import tqdm

from fairchem.core.datasets.atomic_data import atomicdata_list_to_batch

from ._load_utils import _load_from_config

if TYPE_CHECKING:
    from collections.abc import Mapping

    from fairchem.core.modules.normalization.element_references import LinearReferences


class Normalizer(nn.Module):
    """Normalize/denormalize a tensor and optionally add a atom reference offset."""

    def __init__(
        self,
        mean: float | torch.Tensor = 0.0,
        rmsd: float | torch.Tensor = 1.0,
    ):
        """tensor is taken as a sample to calculate the mean and rmsd"""
        super().__init__()

        if isinstance(mean, float):
            mean = torch.tensor(mean)
        if isinstance(rmsd, float):
            rmsd = torch.tensor(rmsd)

        self.register_buffer(name="mean", tensor=mean)
        self.register_buffer(name="rmsd", tensor=rmsd)

    @torch.autocast(device_type="cuda", enabled=False)
    def norm(self, tensor: torch.Tensor) -> torch.Tensor:
        return (tensor - self.mean) / self.rmsd

    @torch.autocast(device_type="cuda", enabled=False)
    def denorm(self, normed_tensor: torch.Tensor) -> torch.Tensor:
        return normed_tensor * self.rmsd + self.mean

    def forward(self, normed_tensor: torch.Tensor) -> torch.Tensor:
        return self.denorm(normed_tensor)

    def load_state_dict(
        self, state_dict: Mapping[str, Any], strict: bool = True, assign: bool = False
    ):
        # check if state dict is legacy state dicts
        if "std" in state_dict:
            state_dict = {
                "mean": torch.tensor(state_dict["mean"]),
                "rmsd": state_dict["std"],
            }

        return super().load_state_dict(state_dict, strict=strict, assign=assign)


def create_normalizer(
    file: str | Path | None = None,
    state_dict: dict | None = None,
    tensor: torch.Tensor | None = None,
    mean: float | torch.Tensor | None = None,
    rmsd: float | torch.Tensor | None = None,
    stdev: float | torch.Tensor | None = None,
) -> Normalizer:
    """Build a target data normalizers with optional atom ref

    Only one of file, state_dict, tensor, or (mean and rmsd) will be used to create a normalizer.
    If more than one set of inputs are given priority will be given following the order in which they are listed above.

    Args:
        file (str or Path): path to pt or npz file.
        state_dict (dict): a state dict for Normalizer module
        tensor (Tensor): a tensor with target values used to compute mean and std
        mean (float | Tensor): mean of target data
        rmsd (float | Tensor): rmsd of target data, rmsd from mean = stdev, rmsd from 0 = rms
        stdev: standard deviation (deprecated, use rmsd instead)

    Returns:
        Normalizer
    """
    if stdev is not None:
        warnings.warn(
            "Use of 'stdev' is deprecated, use 'rmsd' instead", DeprecationWarning
        )
        if rmsd is not None:
            logging.warning(
                "Both 'stdev' and 'rmsd' values where given to create a normalizer, rmsd values will be used."
            )

    # old configs called it stdev, using this in the function signature reduces overhead code elsewhere
    if stdev is not None and rmsd is None:
        rmsd = stdev

    # path takes priority if given
    if file is not None:
        if state_dict is not None or tensor is not None or mean is not None:
            logging.warning(
                "A file to a normalizer has been given. Normalization values will be read from it, and all other inputs"
                " will be ignored."
            )
        extension = Path(file).suffix
        if extension == ".pt":
            # try to load a pt file
            state_dict = torch.load(file)
        elif extension == ".npz":
            # try to load an NPZ file
            values = np.load(file)
            mean = values.get("mean")
            rmsd = values.get("rmsd") or values.get("std")  # legacy files
            tensor = None  # set to None since values read from file are prioritized
        else:
            raise RuntimeError(
                f"Normalizer file with extension '{extension}' is not supported."
            )

    # state dict is second priority
    if state_dict is not None:
        if tensor is not None or mean is not None:
            logging.warning(
                "The state_dict provided will be used to set normalization values. All other inputs will be ignored."
            )
        normalizer = Normalizer()
        normalizer.load_state_dict(state_dict)
        return normalizer

    # if not then read target value tensor
    if tensor is not None:
        if mean is not None:
            logging.warning(
                "Normalization values will be computed from input tensor, all other inputs will be ignored."
            )
        mean = torch.mean(tensor)
        rmsd = torch.std(tensor)
    elif mean is not None and rmsd is not None:
        if not isinstance(mean, torch.Tensor):
            mean = torch.tensor(mean)
        if not isinstance(rmsd, torch.Tensor):
            rmsd = torch.tensor(rmsd)

    # if mean and rmsd are still None than raise an error
    if mean is None or rmsd is None:
        raise ValueError(
            "Incorrect inputs. One of the following sets of inputs must be given: ",
            "a file path to a .pt or .npz file, or mean and rmsd values, or a tensor of target values",
        )

    return Normalizer(mean=mean, rmsd=rmsd)


@torch.no_grad()
def fit_normalizers(
    targets: list[str],
    dataset: Dataset,
    batch_size: int,
    override_values: dict[str, dict[str, float]] | None = None,
    rmsd_correction: int | None = None,
    element_references: dict | None = None,
    num_batches: int | None = None,
    num_workers: int = 0,
    shuffle: bool = True,
    seed: int = 0,
) -> dict[str, Normalizer]:
    """Estimate mean and rmsd from data to create normalizers

    Args:
        targets: list of target names
        dataset: data set to fit linear references with
        batch_size: size of batch
        override_values: dictionary with target names and values to override. i.e. {"forces": {"mean": 0.0}} will set
            the forces mean to zero.
        rmsd_correction: correction to use when computing mean in std/rmsd. See docs for torch.std.
            If not given, will always use 0 when mean == 0, and 1 otherwise.
        element_references:
        num_batches: number of batches to use in fit. If not given will use all batches
        num_workers: number of workers to use in data loader
            Note setting num_workers > 1 leads to finicky multiprocessing issues when using this function
            in distributed mode. The issue has to do with pickling the functions in load_normalizers_from_config
            see function below...
        shuffle: whether to shuffle when loading the dataset
        seed: random seed used to shuffle the sampler if shuffle=True

    Returns:
        dict of normalizer objects
    """
    data_loader = DataLoader(
        dataset,
        batch_size=batch_size,
        shuffle=shuffle,
        collate_fn=atomicdata_list_to_batch,
        num_workers=num_workers,
        persistent_workers=num_workers > 0,
        generator=torch.Generator().manual_seed(seed),
    )

    num_batches = num_batches if num_batches is not None else len(data_loader)
    if num_batches > len(data_loader):
        logging.warning(
            f"The given num_batches {num_batches} is larger than total batches of size {batch_size} in dataset. "
            f"num_batches will be ignored and the whole dataset will be used."
        )
        num_batches = len(data_loader)

    element_references = element_references or {}
    target_vectors = defaultdict(list)

    logging.info(
        f"Estimating mean and rmsd for normalization using {num_batches * batch_size} samples in {num_batches} batches "
        f"of size {batch_size}."
    )
    for i, batch in tqdm(
        enumerate(data_loader), total=num_batches, desc="Estimating mean and rmsd"
    ):
        if i == num_batches:
            break

        for target in targets:
            target_vector = batch[target]
            if target in element_references:
                target_vector = element_references[target].dereference(
                    target_vector, batch, reshaped=False
                )
            target_vectors[target].append(target_vector)

    normalizers = {}
    for target in targets:
        target_vector = torch.cat(target_vectors[target], dim=0)
        values = {"mean": target_vector.mean()}
        if target in override_values:
            for name, val in override_values[target].items():
                values[name] = torch.tensor(val)
        # calculate root mean square deviation
        if "rmsd" not in values:
            if rmsd_correction is None:
                rmsd_correction = 0 if values["mean"] == 0.0 else 1
            values["rmsd"] = (
                ((target_vector - values["mean"]) ** 2).sum()
                / max(len(target_vector) - rmsd_correction, 1)
            ).sqrt()
        normalizers[target] = create_normalizer(**values)

    return normalizers


def load_normalizers_from_config(
    config: dict[str, Any],
    dataset: Dataset,
    seed: int = 0,
    checkpoint_dir: str | Path | None = None,
    element_references: dict[str, LinearReferences] | None = None,
) -> dict[str, Normalizer]:
    """Create a dictionary with element references from a config."""
    # edit the config slightly to extract override args
    if "fit" in config:  # noqa
        if "override_values" not in config["fit"] and isinstance(
            config["fit"]["targets"], dict
        ):
            override_values = {
                target: vals
                for target, vals in config["fit"]["targets"].items()
                if isinstance(vals, dict)
            }
            config["fit"]["override_values"] = override_values
            config["fit"]["targets"] = list(config["fit"]["targets"].keys())

    return _load_from_config(
        config,
        "normalizers",
        fit_normalizers,
        create_normalizer,
        dataset,
        checkpoint_dir,
        seed=seed,
        element_references=element_references,
    )



"""
Modified from https://github.com/pytorch/vision/blob/main/torchvision/ops/stochastic_depth.py
"""

from __future__ import annotations

import torch
from torch import nn


def stochastic_depth_2d(
    input: torch.Tensor, batch: torch.Tensor, p: float, training: bool = True
) -> torch.Tensor:
    """
    Implements the Stochastic Depth from `"Deep Networks with Stochastic Depth"
    <https://arxiv.org/abs/1603.09382>`_ used for randomly dropping residual
    branches of residual architectures.

    Args:
        input (Tensor[num_nodes, ...]): The input tensor or arbitrary dimensions
            with the first one being its node dimension.
        batch (LongTensor[num_nodes]): The batch tensor of the input tensor.
        p (float): probability of the input to be zeroed.
        training: apply stochastic depth if is ``True``. Default: ``True``

    Returns:
        Tensor[N, ...]: The randomly zeroed tensor.
    """
    if p < 0.0 or p > 1.0:
        raise ValueError(f"drop probability has to be between 0 and 1, but got {p}")
    if not training or p == 0.0:
        return input

    batch_size = batch.max() + 1
    survival_rate = 1.0 - p
    size = [batch_size] + [1] * (input.ndim - 1)
    noise = torch.empty(size, dtype=input.dtype, device=input.device)  # type: ignore
    noise = noise.bernoulli_(survival_rate)
    if survival_rate > 0.0:
        noise.div_(survival_rate)
    return input * noise[batch]


# Exact same copy to make torch compile happy, or it will recompile...
def stochastic_depth_3d(
    input: torch.Tensor, batch: torch.Tensor, p: float, training: bool = True
) -> torch.Tensor:
    if p < 0.0 or p > 1.0:
        raise ValueError(f"drop probability has to be between 0 and 1, but got {p}")
    if not training or p == 0.0:
        return input

    batch_size = batch.max() + 1
    survival_rate = 1.0 - p
    size = [batch_size] + [1] * (input.ndim - 1)
    noise = torch.empty(size, dtype=input.dtype, device=input.device)  # type: ignore
    noise = noise.bernoulli_(survival_rate)
    if survival_rate > 0.0:
        noise.div_(survival_rate)
    return input * noise[batch]


class StochasticDepth(nn.Module):
    """
    Stochastic Depth for graph features.
    """

    def __init__(self, p: float) -> None:
        super().__init__()
        self.p = p

    def forward(self, node_features, edge_features, node_batch):
        # Ensure contiguous layout to make torch compile happy
        node_features = node_features.contiguous()
        edge_features = edge_features.contiguous()

        node_features = stochastic_depth_2d(
            input=node_features, batch=node_batch, p=self.p, training=self.training
        )
        edge_features = stochastic_depth_3d(
            input=edge_features, batch=node_batch, p=self.p, training=self.training
        )
        return node_features, edge_features

    def __repr__(self) -> str:
        s = f"{self.__class__.__name__}(p={self.p})"
        return s


class SkipStochasticDepth(nn.Module):
    """
    Skip Stochastic Depth for graph features.
    """

    def forward(self, node_features, edge_features, _):
        return node_features, edge_features

    def __repr__(self) -> str:
        s = f"{self.__class__.__name__}(p=0.0)"
        return s



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.

This script provides the functionality to generate metadata.npz files necessary
for load_balancing the DataLoader.

"""

from __future__ import annotations

import argparse
import multiprocessing as mp
import os
import warnings

import numpy as np
from tqdm import tqdm

from fairchem.core.common.typing import assert_is_instance
from fairchem.core.datasets.collaters.simple_collater import LmdbDataset


def get_data(index):
    data = dataset[index]
    natoms = data.natoms
    neighbors = None
    if hasattr(data, "edge_index") and data.edge_index is not None:
        neighbors = data.edge_index.shape[1]

    return index, natoms, neighbors


def make_lmdb_sizes(args) -> None:
    path = assert_is_instance(args.data_path, str)
    global dataset
    dataset = LmdbDataset({"src": path})
    if os.path.isdir(path):
        outpath = os.path.join(path, "metadata.npz")
    elif os.path.isfile(path):
        outpath = os.path.join(os.path.dirname(path), "metadata.npz")

    output_indices = range(len(dataset))

    pool = mp.Pool(assert_is_instance(args.num_workers, int))
    outputs = list(tqdm(pool.imap(get_data, output_indices), total=len(output_indices)))

    indices = []
    natoms = []
    neighbors = []
    for i in outputs:
        indices.append(i[0])
        natoms.append(i[1])
        neighbors.append(i[2])

    _sort = np.argsort(indices)
    sorted_natoms = np.array(natoms, dtype=np.int32)[_sort]
    if None in neighbors:
        warnings.warn(
            f"edge_index information not found, {outpath} only supports atom-wise load balancing."
        )
        np.savez(outpath, natoms=sorted_natoms)
    else:
        sorted_neighbors = np.array(neighbors, dtype=np.int32)[_sort]
        np.savez(outpath, natoms=sorted_natoms, neighbors=sorted_neighbors)


def get_lmdb_sizes_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--data-path",
        required=True,
        type=str,
        help="Path to S2EF directory or IS2R* .lmdb file",
    )
    parser.add_argument(
        "--num-workers",
        default=1,
        type=int,
        help="Num of workers to parallelize across",
    )
    return parser


if __name__ == "__main__":
    parser = get_lmdb_sizes_parser()
    args: argparse.Namespace = parser.parse_args()
    make_lmdb_sizes(args)



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

import math

import torch
from e3nn.o3 import FromS2Grid, ToS2Grid


class CoefficientMapping(torch.nn.Module):
    """
    Helper module for coefficients used to reshape l <--> m and to get coefficients of specific degree or order

    Args:
        lmax_list (list:int):   List of maximum degree of the spherical harmonics
        mmax_list (list:int):   List of maximum order of the spherical harmonics
        use_rotate_inv_rescale (bool):  Whether to pre-compute inverse rotation rescale matrices
    """

    def __init__(
        self,
        lmax,
        mmax,
    ):
        super().__init__()

        self.lmax = lmax
        self.mmax = mmax
        # Compute the degree (l) and order (m) for each entry of the embedding
        l_harmonic = torch.tensor([]).long()
        m_harmonic = torch.tensor([]).long()
        m_complex = torch.tensor([]).long()

        for l in range(self.lmax + 1):
            mmax = min(self.mmax, l)
            m = torch.arange(-mmax, mmax + 1).long()
            m_complex = torch.cat([m_complex, m], dim=0)
            m_harmonic = torch.cat([m_harmonic, torch.abs(m).long()], dim=0)
            l_harmonic = torch.cat([l_harmonic, m.fill_(l).long()], dim=0)
        self.res_size = len(l_harmonic)

        num_coefficients = len(l_harmonic)
        # `self.to_m` moves m components from different L to contiguous index
        to_m = torch.zeros([num_coefficients, num_coefficients])
        self.m_size = torch.zeros([self.mmax + 1]).long().tolist()

        offset = 0
        for m in range(self.mmax + 1):
            idx_r, idx_i = self.complex_idx(m, -1, m_complex, l_harmonic)

            for idx_out, idx_in in enumerate(idx_r):
                to_m[idx_out + offset, idx_in] = 1.0
            offset = offset + len(idx_r)

            self.m_size[m] = len(idx_r)

            for idx_out, idx_in in enumerate(idx_i):
                to_m[idx_out + offset, idx_in] = 1.0
            offset = offset + len(idx_i)

        to_m = to_m.detach()

        # save tensors and they will be moved to GPU
        self.register_buffer("l_harmonic", l_harmonic, persistent=False)
        self.register_buffer("m_harmonic", m_harmonic, persistent=False)
        self.register_buffer("m_complex", m_complex, persistent=False)
        self.register_buffer("to_m", to_m, persistent=False)

        self.pre_compute_coefficient_idx()

    # Return mask containing coefficients of order m (real and imaginary parts)
    def complex_idx(self, m, lmax, m_complex, l_harmonic):
        """
        Add `m_complex` and `l_harmonic` to the input arguments
        since we cannot use `self.m_complex`.
        """
        if lmax == -1:
            lmax = self.lmax

        indices = torch.arange(len(l_harmonic))
        # Real part
        mask_r = torch.bitwise_and(l_harmonic.le(lmax), m_complex.eq(m))
        mask_idx_r = torch.masked_select(indices, mask_r)

        mask_idx_i = torch.tensor([]).long()
        # Imaginary part
        if m != 0:
            mask_i = torch.bitwise_and(l_harmonic.le(lmax), m_complex.eq(-m))
            mask_idx_i = torch.masked_select(indices, mask_i)

        return mask_idx_r, mask_idx_i

    def pre_compute_coefficient_idx(self):
        """
        Pre-compute the results of `coefficient_idx()` and access them with `prepare_coefficient_idx()`
        """
        lmax = self.lmax
        for l in range(lmax + 1):
            for m in range(lmax + 1):
                mask = torch.bitwise_and(self.l_harmonic.le(l), self.m_harmonic.le(m))
                indices = torch.arange(len(mask))
                mask_indices = torch.masked_select(indices, mask)
                self.register_buffer(
                    f"coefficient_idx_l{l}_m{m}", mask_indices, persistent=False
                )

    def prepare_coefficient_idx(self):
        """
        Construct a list of buffers
        """
        lmax = self.lmax
        coefficient_idx_list = []
        for l in range(lmax + 1):
            l_list = []
            for m in range(lmax + 1):
                l_list.append(getattr(self, f"coefficient_idx_l{l}_m{m}", None))
            coefficient_idx_list.append(l_list)
        return coefficient_idx_list

    # Return mask containing coefficients less than or equal to degree (l) and order (m)
    def coefficient_idx(self, lmax: int, mmax: int):
        if lmax > self.lmax or mmax > self.lmax:
            mask = torch.bitwise_and(self.l_harmonic.le(lmax), self.m_harmonic.le(mmax))
            indices = torch.arange(len(mask), device=mask.device)
            return torch.masked_select(indices, mask)
        else:
            temp = self.prepare_coefficient_idx()
            return temp[lmax][mmax]

    def pre_compute_rotate_inv_rescale(self):
        lmax = self.lmax
        for l in range(lmax + 1):
            for m in range(lmax + 1):
                mask_indices = self.coefficient_idx(l, m)
                rotate_inv_rescale = torch.ones(
                    (1, int((l + 1) ** 2), int((l + 1) ** 2))
                )
                for l_sub in range(l + 1):
                    if l_sub <= m:
                        continue
                    start_idx = l_sub**2
                    length = 2 * l_sub + 1
                    rescale_factor = math.sqrt(length / (2 * m + 1))
                    rotate_inv_rescale[
                        :,
                        start_idx : (start_idx + length),
                        start_idx : (start_idx + length),
                    ] = rescale_factor
                rotate_inv_rescale = rotate_inv_rescale[:, :, mask_indices]
                self.register_buffer(
                    f"rotate_inv_rescale_l{l}_m{m}",
                    rotate_inv_rescale,
                    persistent=False,
                )

    def __repr__(self):
        return f"{self.__class__.__name__}(lmax={self.lmax}, mmax={self.mmax})"


class SO3_Grid(torch.nn.Module):
    """
    Helper functions for grid representation of the irreps

    Args:
        lmax (int):   Maximum degree of the spherical harmonics
        mmax (int):   Maximum order of the spherical harmonics
    """

    def __init__(
        self,
        lmax: int,
        mmax: int,
        normalization: str = "integral",
        resolution: int | None = None,
        rescale: bool = True,
    ):
        super().__init__()
        self.lmax = lmax
        self.mmax = mmax
        self.lat_resolution = 2 * (self.lmax + 1)
        if lmax == mmax:
            self.long_resolution = 2 * (self.mmax + 1) + 1
        else:
            self.long_resolution = 2 * (self.mmax) + 1
        if resolution is not None:
            self.lat_resolution = resolution
            self.long_resolution = resolution

        self.mapping = CoefficientMapping(self.lmax, self.lmax)
        self.rescale = rescale

        to_grid = ToS2Grid(
            self.lmax,
            (self.lat_resolution, self.long_resolution),
            normalization=normalization,  # normalization="integral",
        )
        to_grid_mat = torch.einsum("mbi, am -> bai", to_grid.shb, to_grid.sha).detach()
        # rescale based on mmax
        if rescale and lmax != mmax:
            for lval in range(lmax + 1):
                if lval <= mmax:
                    continue
                start_idx = lval**2
                length = 2 * lval + 1
                rescale_factor = math.sqrt(length / (2 * mmax + 1))
                to_grid_mat[:, :, start_idx : (start_idx + length)] = (
                    to_grid_mat[:, :, start_idx : (start_idx + length)] * rescale_factor
                )
        to_grid_mat = to_grid_mat[
            :, :, self.mapping.coefficient_idx(self.lmax, self.mmax)
        ]

        from_grid = FromS2Grid(
            (self.lat_resolution, self.long_resolution),
            self.lmax,
            normalization=normalization,  # normalization="integral",
        )
        from_grid_mat = torch.einsum(
            "am, mbi -> bai", from_grid.sha, from_grid.shb
        ).detach()
        # rescale based on mmax
        if rescale and lmax != mmax:
            for lval in range(lmax + 1):
                if lval <= mmax:
                    continue
                start_idx = lval**2
                length = 2 * lval + 1
                rescale_factor = math.sqrt(length / (2 * mmax + 1))
                from_grid_mat[:, :, start_idx : (start_idx + length)] = (
                    from_grid_mat[:, :, start_idx : (start_idx + length)]
                    * rescale_factor
                )
        from_grid_mat = from_grid_mat[
            :, :, self.mapping.coefficient_idx(self.lmax, self.mmax)
        ]

        # save tensors and they will be moved to GPU
        self.register_buffer("to_grid_mat", to_grid_mat, persistent=False)
        self.register_buffer("from_grid_mat", from_grid_mat, persistent=False)

    # Compute matrices to transform irreps to grid
    def get_to_grid_mat(self, device=None):
        return self.to_grid_mat

    # Compute matrices to transform grid to irreps
    def get_from_grid_mat(self, device=None):
        return self.from_grid_mat

    # Compute grid from irreps representation
    def to_grid(self, embedding, lmax: int, mmax: int):
        to_grid_mat = self.to_grid_mat[:, :, self.mapping.coefficient_idx(lmax, mmax)]
        return torch.einsum("bai, zic -> zbac", to_grid_mat, embedding)

    # Compute irreps from grid representation
    def from_grid(self, grid, lmax: int, mmax: int):
        from_grid_mat = self.from_grid_mat[
            :, :, self.mapping.coefficient_idx(lmax, mmax)
        ]
        return torch.einsum("bai, zbac -> zic", from_grid_mat, grid)



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

import copy
import logging
import multiprocessing as mp
import os
import random
from collections import defaultdict
from contextlib import nullcontext
from functools import wraps
from typing import TYPE_CHECKING, Protocol, Sequence

import hydra
import numpy as np
import torch
from torch.distributed.elastic.utils.distributed import get_free_port
from torchtnt.framework import PredictUnit, State

from fairchem.core.common import gp_utils
from fairchem.core.common.distutils import (
    CURRENT_DEVICE_TYPE_STR,
    get_device_for_local_rank,
)
from fairchem.core.datasets.atomic_data import AtomicData
from fairchem.core.units.mlip_unit import InferenceSettings
from fairchem.core.units.mlip_unit.inference.client_websocket import (
    SyncMLIPInferenceWebSocketClient,
)
from fairchem.core.units.mlip_unit.inference.inference_server_ray import (
    MLIPInferenceServerWebSocket,
)
from fairchem.core.units.mlip_unit.utils import (
    load_inference_model,
    tf32_context_manager,
)

if TYPE_CHECKING:
    from fairchem.core.units.mlip_unit.mlip_unit import Task


def collate_predictions(predict_fn):
    @wraps(predict_fn)
    def collated_predict(
        predict_unit, data: AtomicData, undo_element_references: bool = True
    ):
        # Get the full prediction dictionary from the original predict method
        preds = predict_fn(predict_unit, data, undo_element_references)
        if gp_utils.initialized():
            data.batch = data.batch_full
        collated_preds = defaultdict(list)
        for i, dataset in enumerate(data.dataset):
            for task in predict_unit.dataset_to_tasks[dataset]:
                if task.level == "system":
                    collated_preds[task.property].append(
                        preds[task.name][i].unsqueeze(0)
                    )
                elif task.level == "atom":
                    collated_preds[task.property].append(
                        preds[task.name][data.batch == i]
                    )
                else:
                    raise RuntimeError(
                        f"Unrecognized task level={task.level} found in data batch at position {i}"
                    )

        return {prop: torch.cat(val) for prop, val in collated_preds.items()}

    return collated_predict


class MLIPPredictUnitProtocol(Protocol):
    def predict(self, data: AtomicData, undo_element_references: bool) -> dict: ...

    @property
    def dataset_to_tasks(self) -> dict[str, list]: ...


class MLIPPredictUnit(PredictUnit[AtomicData], MLIPPredictUnitProtocol):
    def __init__(
        self,
        inference_model_path: str,
        device: str = "cpu",
        overrides: dict | None = None,
        inference_settings: InferenceSettings | None = None,
        seed: int = 41,
        atom_refs: dict | None = None,
        assert_on_nans: bool = False,
    ):
        super().__init__()
        os.environ[CURRENT_DEVICE_TYPE_STR] = device

        self.set_seed(seed)
        # note these are different from the element references used for model training
        self.atom_refs = (
            {task.replace("_elem_refs", ""): refs for task, refs in atom_refs.items()}
            if atom_refs is not None
            else {}
        )

        if inference_settings is None:
            inference_settings = InferenceSettings()
        if overrides is None:
            overrides = {}
        if "backbone" not in overrides:
            overrides["backbone"] = {}
        # always disable always_use_pbc for inference
        overrides["backbone"]["always_use_pbc"] = False
        if inference_settings.activation_checkpointing is not None:
            overrides["backbone"]["activation_checkpointing"] = (
                inference_settings.activation_checkpointing
            )
        if inference_settings.wigner_cuda is not None:
            overrides["backbone"]["use_cuda_graph_wigner"] = (
                inference_settings.wigner_cuda
            )
        if inference_settings.external_graph_gen is not None:
            overrides["backbone"][
                "otf_graph"
            ] = not inference_settings.external_graph_gen

        if inference_settings.internal_graph_gen_version is not None:
            overrides["backbone"]["radius_pbc_version"] = (
                inference_settings.internal_graph_gen_version
            )

        self.model, checkpoint = load_inference_model(
            inference_model_path, use_ema=True, overrides=overrides
        )
        tasks = [
            hydra.utils.instantiate(task_config)
            for task_config in checkpoint.tasks_config
        ]
        self.tasks = {t.name: t for t in tasks}

        self._dataset_to_tasks = get_dataset_to_tasks_map(self.tasks.values())
        assert set(self._dataset_to_tasks.keys()).issubset(
            set(self.model.module.backbone.dataset_list)
        ), "Datasets in tasks is not a strict subset of datasets in backbone."
        assert device in ["cpu", "cuda"], "device must be either 'cpu' or 'cuda'"

        self.device = get_device_for_local_rank() if device == "cuda" else "cpu"

        self.model.eval()

        self.lazy_model_intialized = False
        self.inference_mode = inference_settings

        # store composition embedding of system the model was merged on
        self.merged_on = None

        self.assert_on_nans = assert_on_nans

        if self.direct_forces:
            logging.warning(
                "This is a direct-force model. Direct force predictions may lead to discontinuities in the potential "
                "energy surface and energy conservation errors."
            )

    @property
    def direct_forces(self) -> bool:
        return self.model.module.backbone.direct_forces

    @property
    def dataset_to_tasks(self) -> dict[str, list]:
        return self._dataset_to_tasks

    def set_seed(self, seed: int):
        logging.debug(f"Setting random seed to {seed}")
        self._seed = seed
        random.seed(seed)
        np.random.seed(seed)
        torch.manual_seed(seed)
        torch.cuda.manual_seed_all(seed)

    def move_to_device(self):
        self.model.to(self.device)
        for task in self.tasks.values():
            task.normalizer.to(self.device)
            if task.element_references is not None:
                task.element_references.to(self.device)

    def predict_step(self, state: State, data: AtomicData) -> dict[str, torch.tensor]:
        return self.predict(data)

    def get_composition_charge_spin_dataset(self, data):
        composition_sum = data.atomic_numbers.new_zeros(
            self.model.module.backbone.max_num_elements,
            dtype=torch.int,
        ).index_add(
            0,
            data.atomic_numbers.to(torch.int),
            data.atomic_numbers.new_ones(data.atomic_numbers.shape[0], dtype=torch.int),
        )
        comp_charge_spin = (
            composition_sum,
            getattr(data, "charge", None),
            getattr(data, "spin", None),
        )
        return comp_charge_spin, getattr(data, "dataset", [None])

    @collate_predictions
    def predict(
        self, data: AtomicData, undo_element_references: bool = True
    ) -> dict[str, torch.tensor]:
        if not self.lazy_model_intialized:
            # merge everything on CPU
            if self.inference_mode.merge_mole:
                # replace backbone with non MOE version
                assert (
                    data.natoms.numel() == 1
                ), f"Cannot merge model with multiple systems in batch. Must be exactly 1 system, found {data.natoms.numel()}"
                self.model.module.backbone = (
                    self.model.module.backbone.merge_MOLE_model(data.clone())
                )
                self.model.eval()
            # move to device
            self.move_to_device()
            if self.inference_mode.compile:
                logging.warning(
                    "Model is being compiled this might take a while for the first time"
                )
                self.model = torch.compile(self.model, dynamic=True)
            self.lazy_model_intialized = True

        if self.inference_mode.external_graph_gen and data.edge_index.shape[1] == 0:
            raise ValueError(
                "Cannot run inference with external graph generation on empty edge index. "
                "Please ensure the input data has valid edges."
            )

        data_device = data.to(self.device)

        if self.inference_mode.merge_mole:
            if self.merged_on is None:
                # only get embeddings after moved to final device to get right types
                self.merged_on = self.get_composition_charge_spin_dataset(data_device)
            else:
                this_sys = self.get_composition_charge_spin_dataset(data_device)
                assert (
                    data_device.natoms.numel() == 1
                ), f"Cannot run merged model on batch with multiple systems. Must be exactly 1 system, found {data_device.natoms.numel()}"
                assert (
                    self.merged_on[0][0].isclose(this_sys[0][0], rtol=1e-5).all()
                ), "Cannot run on merged model on system. Embeddings seem different..."
                assert (
                    self.merged_on[0][1] == this_sys[0][1]
                ), f"Cannot run on merged model on system. Charge is diferrent {self.merged_on[0][1]} vs {this_sys[0][1]}"
                assert (
                    self.merged_on[0][2] == this_sys[0][2]
                ), f"Cannot run on merged model on system. Spin is diferrent {self.merged_on[0][2]} vs {this_sys[0][2]}"
                assert (
                    self.merged_on[1] == this_sys[1]
                ), f"Cannot run on merged model on system. Dataset is diferrent {self.merged_on[1]} vs {this_sys[1]}"

        inference_context = torch.no_grad() if self.direct_forces else nullcontext()
        tf32_context = (
            tf32_context_manager() if self.inference_mode.tf32 else nullcontext()
        )

        pred_output = {}
        with inference_context, tf32_context:
            output = self.model(data_device)
            for task_name, task in self.tasks.items():
                pred_output[task_name] = task.normalizer.denorm(
                    output[task_name][task.property]
                )
                if self.assert_on_nans:
                    assert torch.isfinite(
                        pred_output[task_name]
                    ).all(), f"NaNs/Infs found in prediction for task {task_name}.{task.property}"
                if undo_element_references and task.element_references is not None:
                    pred_output[task_name] = task.element_references.undo_refs(
                        data_device, pred_output[task_name]
                    )

        return pred_output


def get_dataset_to_tasks_map(tasks: Sequence[Task]) -> dict[str, list[Task]]:
    """Create a mapping from dataset names to their associated tasks.

    Args:
        tasks: A sequence of Task objects to be organized by dataset

    Returns:
        A dictionary mapping dataset names (str) to lists of Task objects
        that are associated with that dataset
    """
    dset_to_tasks_map = defaultdict(list)
    for task in tasks:
        for dataset_name in task.datasets:
            dset_to_tasks_map[dataset_name].append(task)
    return dict(dset_to_tasks_map)


def _run_server_process(predictor_config, port, num_workers, ready_queue):
    """Function to run server in separate process"""
    try:
        server = MLIPInferenceServerWebSocket(
            predictor_config=predictor_config,
            port=port,
            num_workers=num_workers,
        )
        # Signal that server is ready
        ready_queue.put("ready")
        server.run()
    except Exception as e:
        ready_queue.put(f"error: {e}")


class ParallelMLIPPredictUnit(MLIPPredictUnitProtocol):
    def __init__(
        self,
        inference_model_path: str,
        device: str = "cpu",
        overrides: dict | None = None,
        inference_settings: InferenceSettings | None = None,
        seed: int = 41,
        atom_refs: dict | None = None,
        assert_on_nans: bool = False,
        server_config: dict | None = None,
        client_config: dict | None = None,
    ):
        """
        This PredictUnit can be used to run inference on a remote server.

        It can be used in several modes:
        1) If server_config is provided then it will start a local server in a thread and create a client to connect to it.
        A separate client cannot be provided in this case.
        2) If server_config is NOT provided and client_config is provided, we assume the remote server is already running
        and we will create a client to connect to it.
        """
        super().__init__()
        assert (server_config is not None) ^ (
            client_config is not None
        ), "Exactly one of server_config or client_config must be provided."

        config = {}
        self.server_process = None

        # TODO, we need this just to get the datasets for the FAIRChemCalculator, this is not great, think about if
        # we can remove this dependency
        _mlip_pred_unit = MLIPPredictUnit(
            inference_model_path=inference_model_path,
            device="cpu",
            overrides=overrides,
            inference_settings=inference_settings,
            seed=seed,
            atom_refs=atom_refs,
        )
        self._dataset_to_tasks = copy.deepcopy(_mlip_pred_unit.dataset_to_tasks)

        if server_config is not None:
            logging.info(f"Starting inference server with config {server_config}")
            if "port" not in server_config:
                server_config["port"] = get_free_port()
            config["server"] = server_config
            self.server_address = "localhost"
            self.server_port = server_config.get("port")
            self.workers = server_config.get("workers", 1)
            predict_unit_config = {
                "_target_": "fairchem.core.units.mlip_unit.predict.MLIPPredictUnit",
                "inference_model_path": inference_model_path,
                "device": device,
                "overrides": overrides,
                "inference_settings": inference_settings,
                "seed": seed,
                "atom_refs": atom_refs,
                "assert_on_nans": assert_on_nans,
            }

            self._start_server_process(
                predict_unit_config, self.server_port, self.workers
            )

        if client_config is not None:
            logging.info(f"Connecting to inference server with config {client_config}")
            self.client = hydra.utils.instantiate(client_config)
        else:
            self.client = SyncMLIPInferenceWebSocketClient(
                host=self.server_address,
                port=self.server_port,
            )

    def _start_server_process(self, predict_unit_config, port, workers):
        """Start server process and wait for it to be ready"""
        # Create a queue to check server readiness
        self.ready_queue = mp.Queue()

        # Start server in separate process instead of thread
        self.server_process = mp.Process(
            target=_run_server_process,
            args=(
                predict_unit_config,
                port,
                workers,
                self.ready_queue,
            ),
        )
        self.server_process.start()

        # Wait for server to be ready (with timeout)
        try:
            result = self.ready_queue.get(timeout=30)  # 30 second timeout
            if result != "ready":
                raise RuntimeError(f"Server failed to start: {result}")
            logging.info("Server is ready")
        except Exception as e:
            if self.server_process.is_alive():
                self.server_process.terminate()
            raise e

    def cleanup(self):
        logging.info("Shutting down ParallelMLIPPredictUnit")
        # Clean up server process if it was started locally
        if hasattr(self, "server_process") and self.server_process.is_alive():
            self.server_process.terminate()
            self.server_process.join(timeout=10)
            if self.server_process.is_alive():
                self.server_process.kill()

    def __del__(self):
        self.cleanup()

    def predict(
        self, data: AtomicData, undo_element_references: bool = True
    ) -> dict[str, torch.tensor]:
        """
        Predict method that sends data to the remote server and returns predictions.
        """
        if not hasattr(self, "client"):
            raise RuntimeError(
                "Client is not initialized. Ensure server_config or client_config is provided."
            )
        return self.client.call(data)

    @property
    def dataset_to_tasks(self) -> dict[str, list]:
        return self._dataset_to_tasks



from __future__ import annotations

from typing import TYPE_CHECKING

import torch
from torch import nn

if TYPE_CHECKING:
    from fairchem.core.models.escaip.configs import (
        GlobalConfigs,
        GraphNeuralNetworksConfigs,
        RegularizationConfigs,
    )
from fairchem.core.models.escaip.utils.graph_utils import (
    compilable_scatter,
)
from fairchem.core.models.escaip.utils.nn_utils import (
    Activation,
    NormalizationType,
    get_feedforward,
    get_normalization_layer,
)


class ReadoutBlock(nn.Module):
    """
    Readout from each graph attention block for energy and force output
    """

    def __init__(
        self,
        global_cfg: GlobalConfigs,
        gnn_cfg: GraphNeuralNetworksConfigs,
        reg_cfg: RegularizationConfigs,
    ):
        super().__init__()

        self.backbone_dtype = (
            torch.float16 if global_cfg.use_fp16_backbone else torch.float32
        )

        self.energy_reduce = gnn_cfg.energy_reduce
        self.use_edge_readout = global_cfg.regress_forces and global_cfg.direct_forces
        self.use_global_readout = gnn_cfg.use_global_readout

        # global read out
        if gnn_cfg.use_global_readout:
            self.global_ffn = get_feedforward(
                hidden_dim=global_cfg.hidden_size,
                activation=Activation(global_cfg.activation),
                hidden_layer_multiplier=gnn_cfg.readout_hidden_layer_multiplier,
                dropout=reg_cfg.mlp_dropout,
                bias=True,
            ).to(self.backbone_dtype)
            self.pre_global_norm = get_normalization_layer(
                NormalizationType(reg_cfg.normalization)
            )(global_cfg.hidden_size, dtype=self.backbone_dtype)
            # self.post_global_norm = get_normalization_layer(
            #     NormalizationType(reg_cfg.normalization)
            # )(global_cfg.hidden_size, dtype=self.backbone_dtype)

        # node read out
        self.node_ffn = get_feedforward(
            hidden_dim=global_cfg.hidden_size,
            activation=Activation(global_cfg.activation),
            hidden_layer_multiplier=gnn_cfg.readout_hidden_layer_multiplier,
            dropout=reg_cfg.mlp_dropout,
            bias=True,
        ).to(self.backbone_dtype)
        self.pre_node_norm = get_normalization_layer(
            NormalizationType(reg_cfg.normalization)
        )(global_cfg.hidden_size, dtype=self.backbone_dtype)
        # self.post_node_norm = get_normalization_layer(
        #     NormalizationType(reg_cfg.normalization)
        # )(global_cfg.hidden_size, dtype=self.backbone_dtype)

        # forces read out
        if self.use_edge_readout:
            self.edge_ffn = get_feedforward(
                hidden_dim=global_cfg.hidden_size,
                activation=Activation(global_cfg.activation),
                hidden_layer_multiplier=gnn_cfg.readout_hidden_layer_multiplier,
                dropout=reg_cfg.mlp_dropout,
                bias=True,
            ).to(self.backbone_dtype)
            self.pre_edge_norm = get_normalization_layer(
                NormalizationType(reg_cfg.normalization)
            )(global_cfg.hidden_size, dtype=self.backbone_dtype)
            # self.post_edge_norm = get_normalization_layer(
            #     NormalizationType(reg_cfg.normalization)
            # )(global_cfg.hidden_size, dtype=self.backbone_dtype)

    def forward(self, data, node_features, edge_features):
        """
        Output:
            Global Readout (G, H);
            Node Readout (N, H);
            Edge Readout (N, max_nei, H)
        """
        node_readout = node_features + self.node_ffn(self.pre_node_norm(node_features))
        # node_readout = self.post_node_norm(node_readout)

        if self.use_global_readout:
            global_features = compilable_scatter(
                src=node_features,
                index=data.node_batch,
                dim_size=data.graph_padding_mask.shape[0],
                dim=0,
                reduce=self.energy_reduce,
            )
            global_readout = global_features + self.global_ffn(
                self.pre_global_norm(global_features)
            )
            # global_readout = self.post_global_norm(global_readout)
        else:
            global_readout = torch.zeros_like(node_readout)

        if self.use_edge_readout:
            edge_readout = edge_features + self.edge_ffn(
                self.pre_edge_norm(edge_features)
            )
            # edge_readout = self.post_edge_norm(edge_readout)
        else:
            edge_readout = torch.zeros_like(node_readout)

        return global_readout, node_readout, edge_readout



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

from typing import TypeVar

import pandas as pd

from fairchem.core.components.benchmark.benchmark_reducer import (
    JsonDFReducer,
)

R = TypeVar("R")
M = TypeVar("M")


class AdsorptionReducer(JsonDFReducer):
    def __init__(
        self,
        benchmark_name: str,
        target_data_key: str | None = None,
        index_name: str | None = None,
    ):
        """
        Args:
            benchmark_name: Name of the benchmark, used for file naming
            target_data_key: Key corresponding to the target value in the results
        """
        self.index_name = index_name
        self.benchmark_name = benchmark_name
        self.target_data_key = target_data_key

    def compute_metrics(self, results: pd.DataFrame, run_name: str) -> pd.DataFrame:
        """
        Compute mean absolute error metrics for common columns between results and targets.

        Args:
            results: DataFrame containing prediction results
            run_name: Name of the current run, used as index in the metrics DataFrame

        Returns:
            DataFrame containing computed metrics with run_name as index
        """
        """This will just compute MAE of everything that is common in the results and target dataframes"""
        pred_cols = [col for col in results.columns if col != self.target_data_key]
        targets = results[self.target_data_key]
        metrics = {
            f"{col},mae": (results[col] - targets).abs().mean() for col in pred_cols
        }
        return pd.DataFrame([metrics], index=[run_name])



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.

# Copyright (c) Meta Platforms, Inc. and affiliates.
# Borrowed from https://github.com/facebookresearch/pythia/blob/master/pythia/common/registry.py.

Registry is central source of truth. Inspired from Redux's concept of
global store, Registry maintains mappings of various information to unique
keys. Special functions in registry can be used as decorators to register
different kind of classes.

Import the global registry object using

``from fairchem.core.common.registry import registry``

Various decorators for registry different kind of classes with unique keys

- Register a model: ``@registry.register_model``
"""

from __future__ import annotations

import importlib
from typing import Any, Callable, ClassVar, TypeVar, Union

R = TypeVar("R")
NestedDict = dict[str, Union[str, Callable[..., Any], "NestedDict"]]


def _get_absolute_mapping(name: str):
    # in this case, the `name` should be the fully qualified name of the class
    # e.g., `fairchem.core.tasks.base_task.BaseTask`
    # we can use importlib to get the module (e.g., `fairchem.core.tasks.base_task`)
    # and then import the class (e.g., `BaseTask`)

    module_name = ".".join(name.split(".")[:-1])
    class_name = name.split(".")[-1]

    try:
        module = importlib.import_module(module_name)
    except (ModuleNotFoundError, ValueError) as e:
        raise RuntimeError(
            f"Could not import module `{module_name}` for import `{name}`"
        ) from e

    try:
        return getattr(module, class_name)
    except AttributeError as e:
        raise RuntimeError(
            f"Could not import class `{class_name}` from module `{module_name}`"
        ) from e


class Registry:
    r"""Class for registry object which acts as central source of truth."""

    mapping: ClassVar[NestedDict] = {
        # Mappings to respective classes.
        "task_name_mapping": {},
        "dataset_name_mapping": {},
        "loss_name_mapping": {},
        "model_name_mapping": {},
        "logger_name_mapping": {},
        "trainer_name_mapping": {},
        "state": {},
    }

    @classmethod
    def register_task(cls, name: str):
        r"""Register a new task to registry with key 'name'
        Args:
            name: Key with which the task will be registered.
        Usage::
            from fairchem.core.common.registry import registry
            from fairchem.core.tasks import BaseTask
            @registry.register_task("train")
            class TrainTask(BaseTask):
                ...
        """

        def wrap(func: Callable[..., R]) -> Callable[..., R]:
            cls.mapping["task_name_mapping"][name] = func
            return func

        return wrap

    @classmethod
    def register_dataset(cls, name: str):
        r"""Register a dataset to registry with key 'name'

        Args:
            name: Key with which the dataset will be registered.

        Usage::

            from fairchem.core.common.registry import registry
            from fairchem.core.datasets import BaseDataset

            @registry.register_dataset("qm9")
            class QM9(BaseDataset):
                ...
        """

        def wrap(func: Callable[..., R]) -> Callable[..., R]:
            cls.mapping["dataset_name_mapping"][name] = func
            return func

        return wrap

    @classmethod
    def register_loss(cls, name):
        r"""Register a loss to registry with key 'name'

        Args:
            name: Key with which the loss will be registered.

        Usage::

            from fairchem.core.common.registry import registry
            from torch import nn

            @registry.register_loss("mae")
            class MAELoss(nn.Module):
                ...

        """

        def wrap(func):
            from torch import nn

            assert issubclass(
                func, nn.Module
            ), "All loss must inherit torch.nn.Module class"
            cls.mapping["loss_name_mapping"][name] = func
            return func

        return wrap

    @classmethod
    def register_model(cls, name: str):
        r"""Register a model to registry with key 'name'

        Args:
            name: Key with which the model will be registered.

        Usage::

            from fairchem.core.common.registry import registry
            from fairchem.core.modules.layers import CGCNNConv

            @registry.register_model("cgcnn")
            class CGCNN():
                ...
        """

        def wrap(func: Callable[..., R]) -> Callable[..., R]:
            cls.mapping["model_name_mapping"][name] = func
            return func

        return wrap

    @classmethod
    def register_logger(cls, name: str):
        r"""Register a logger to registry with key 'name'

        Args:
            name: Key with which the logger will be registered.

        Usage::

            from fairchem.core.common.registry import registry

            @registry.register_logger("wandb")
            class WandBLogger():
                ...
        """

        def wrap(func: Callable[..., R]) -> Callable[..., R]:
            from fairchem.core.common.logger import Logger

            assert issubclass(func, Logger), "All loggers must inherit Logger class"
            cls.mapping["logger_name_mapping"][name] = func
            return func

        return wrap

    @classmethod
    def register_trainer(cls, name: str):
        r"""Register a trainer to registry with key 'name'

        Args:
            name: Key with which the trainer will be registered.

        Usage::

            from fairchem.core.common.registry import registry

            @registry.register_trainer("active_discovery")
            class ActiveDiscoveryTrainer():
                ...
        """

        def wrap(func: Callable[..., R]) -> Callable[..., R]:
            cls.mapping["trainer_name_mapping"][name] = func
            return func

        return wrap

    @classmethod
    def register(cls, name: str, obj) -> None:
        r"""Register an item to registry with key 'name'

        Args:
            name: Key with which the item will be registered.

        Usage::

            from fairchem.core.common.registry import registry

            registry.register("config", {})
        """
        path = name.split(".")
        current = cls.mapping["state"]

        for part in path[:-1]:
            if part not in current:
                current[part] = {}
            current = current[part]

        current[path[-1]] = obj

    @classmethod
    def __import_error(cls, name: str, mapping_name: str) -> RuntimeError:
        kind = mapping_name[: -len("_name_mapping")]
        mapping = cls.mapping.get(mapping_name, {})
        existing_keys: list[str] = list(mapping.keys())

        if len(existing_keys) == 0:
            raise RuntimeError(
                f"Registry for {mapping_name} is empty. You may have forgot to load a module."
            )

        existing_cls_path = (
            mapping.get(existing_keys[-1], None) if existing_keys else None
        )
        if existing_cls_path is not None:
            existing_cls_path = (
                f"{existing_cls_path.__module__}.{existing_cls_path.__qualname__}"
            )
        else:
            existing_cls_path = "fairchem.core.trainers.ForcesTrainer"

        existing_keys = [f"'{name}'" for name in existing_keys]
        existing_keys = ", ".join(existing_keys[:-1]) + " or " + existing_keys[-1]
        existing_keys_str = f" (one of {existing_keys})" if existing_keys else ""
        return RuntimeError(
            f"Failed to find the {kind} '{name}'. "
            f"You may either use a {kind} from the registry{existing_keys_str} "
            f"or provide the full import path to the {kind} (e.g., '{existing_cls_path}')."
        )

    @classmethod
    def get_class(cls, name: str, mapping_name: str):
        existing_mapping = cls.mapping[mapping_name].get(name, None)
        if existing_mapping is not None:
            return existing_mapping

        # mapping be class path of type `{module_name}.{class_name}` (e.g., `fairchem.core.trainers.ForcesTrainer`)
        if name.count(".") < 1:
            raise cls.__import_error(name, mapping_name)

        try:
            return _get_absolute_mapping(name)
        except RuntimeError as e:
            raise cls.__import_error(name, mapping_name) from e

    @classmethod
    def get_task_class(cls, name: str):
        return cls.get_class(name, "task_name_mapping")

    @classmethod
    def get_dataset_class(cls, name: str):
        return cls.get_class(name, "dataset_name_mapping")

    @classmethod
    def get_loss_class(cls, name):
        return cls.get_class(name, "loss_name_mapping")

    @classmethod
    def get_model_class(cls, name: str):
        return cls.get_class(name, "model_name_mapping")

    @classmethod
    def get_logger_class(cls, name: str):
        return cls.get_class(name, "logger_name_mapping")

    @classmethod
    def get_trainer_class(cls, name: str):
        return cls.get_class(name, "trainer_name_mapping")

    @classmethod
    def get(cls, name: str, default=None, no_warning: bool = False):
        r"""Get an item from registry with key 'name'

        Args:
            name (string): Key whose value needs to be retrieved.
            default: If passed and key is not in registry, default value will
                     be returned with a warning. Default: None
            no_warning (bool): If passed as True, warning when key doesn't exist
                               will not be generated. Useful for cgcnn's
                               internal operations. Default: False
        Usage::

            from fairchem.core.common.registry import registry

            config = registry.get("config")
        """
        original_name = name
        split_name = name.split(".")
        value = cls.mapping["state"]
        for subname in split_name:
            value = value.get(subname, default)
            if value is default:
                break

        if (
            "writer" in cls.mapping["state"]
            and value == default
            and no_warning is False
        ):
            cls.mapping["state"]["writer"].write(
                f"Key {original_name} is not present in registry, returning default value "
                f"of {default}"
            )
        return value

    @classmethod
    def unregister(cls, name: str):
        r"""Remove an item from registry with key 'name'

        Args:
            name: Key which needs to be removed.
        Usage::

            from fairchem.core.common.registry import registry

            config = registry.unregister("config")
        """
        return cls.mapping["state"].pop(name, None)


registry = Registry()



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

import logging
import os
import traceback
import warnings
from copy import deepcopy
from typing import Any, ClassVar

import ase.io
import ase.units
import numpy as np
import pandas as pd
from ase.constraints import FixSymmetry
from ase.filters import FrechetCellFilter
from ase.optimize import FIRE
from monty.dev import requires

try:
    # pylint: disable=E0611
    from matbench_discovery.phonons import check_imaginary_freqs
    from matbench_discovery.phonons import thermal_conductivity as ltc
    from moyopy import MoyoDataset
    from moyopy.interface import MoyoAdapter
    from pymatviz.enums import Key
    from tqdm import tqdm

    mbd_installed = True
except ImportError:
    mbd_installed = False

from fairchem.core.components.calculate import CalculateRunner


def get_kappa103_data_list(reference_data_path: str, debug=False):
    atoms = ase.io.read(reference_data_path, format="extxyz", index=":")
    if debug:
        atoms = atoms[:1]
    return atoms


@requires(
    mbd_installed,
    message="Requires `matbench_discovery[symmetry,phonons]` to be installed",
)
class KappaRunner(CalculateRunner):
    """Calculate elastic tensor for a set of structures."""

    result_glob_pattern: ClassVar[str] = "kappa103_dist*_*-*.json.gz"

    def __init__(
        self,
        calculator,
        input_data,
        displacement: float = 0.03,
    ):
        """Initialize the CalculateRunner with a calculator and input data.

        Args:
            calculator (Calculator): ASE-like calculator to perform calculations
            input_data (Sequence): Input data to be processed by the calculator
        """
        super().__init__(calculator, input_data)
        self.displacement = displacement

    # TODO continue if unfinished
    def calculate(self, job_num: int = 0, num_jobs: int = 1) -> list[dict[str, Any]]:
        max_steps = 5000
        force_max = 1e-4  # Run until the forces are smaller than this in eV/A
        symprec = 1e-5
        enforce_relax_symm = True
        conductivity_broken_symm = False
        prog_bar = True
        # save_forces = False  # Save force sets to file
        temperatures = [300]

        all_results = []

        chunk_indices = np.array_split(range(len(self.input_data)), num_jobs)[job_num]

        for i in tqdm(chunk_indices, desc="Running kappa calculations"):
            atoms = self.input_data[i]

            mat_id = atoms.info.get(Key.mat_id, f"id-{len(all_results)}")
            init_info = deepcopy(atoms.info)
            formula = atoms.info.get("name", "unknown")
            spg_num = MoyoDataset(MoyoAdapter.from_atoms(atoms)).number

            info_dict = {
                "sid": mat_id,
                Key.mat_id: mat_id,
                Key.formula: formula,
                Key.spg_num: spg_num,
                "errors": [],
                "error_traceback": [],
            }

            relax_dict = {
                "max_stress": None,
                "reached_max_steps": False,
                "broken_symmetry": False,
            }

            try:
                # NOTE: not using the relax_atom recipe because need to use mask for cellfilter.
                atoms.calc = self.calculator
                if max_steps > 0:
                    if enforce_relax_symm:
                        atoms.set_constraint(FixSymmetry(atoms))

                    # Use standard mask for no-tilt constraint
                    filtered_atoms = FrechetCellFilter(
                        atoms, mask=[True] * 3 + [False] * 3
                    )

                    optimizer = FIRE(filtered_atoms, logfile=None)
                    optimizer.run(fmax=force_max, steps=max_steps)

                    reached_max_steps = optimizer.Nsteps >= max_steps
                    if reached_max_steps:
                        logging.info(
                            f"{mat_id=} reached {max_steps=} during relaxation."
                        )

                    max_stress = (
                        atoms.get_stress().reshape((2, 3), order="C").max(axis=1)
                    )
                    atoms.calc = None
                    atoms.constraints = None
                    atoms.info = init_info | atoms.info

                    # Check if symmetry was broken during relaxation
                    relaxed_spg = MoyoDataset(MoyoAdapter.from_atoms(atoms)).number
                    broken_symmetry = spg_num != relaxed_spg
                    relax_dict = {
                        "max_stress": max_stress,
                        "reached_max_steps": reached_max_steps,
                        "relaxed_space_group_number": relaxed_spg,
                        "broken_symmetry": broken_symmetry,
                    }

            except Exception as exc:
                warnings.warn(
                    f"Failed to relax {formula=}, {mat_id=}: {exc!r}", stacklevel=2
                )
                traceback.print_exc()
                info_dict["errors"].append(f"RelaxError: {exc!r}")
                info_dict["error_traceback"].append(traceback.format_exc())
                results = info_dict | relax_dict
                all_results.append(results)
                continue

            # Calculation of force sets
            try:
                # Initialize phono3py with the relaxed structure
                ph3 = ltc.init_phono3py(
                    atoms,
                    fc2_supercell=atoms.info.get("fc2_supercell", [2, 2, 2]),
                    fc3_supercell=atoms.info.get("fc3_supercell", [2, 2, 2]),
                    q_point_mesh=atoms.info.get("q_point_mesh", [10, 10, 10]),
                    displacement_distance=self.displacement,
                    symprec=symprec,
                )

                # Calculate force constants and frequencies
                ph3, fc2_set, freqs = ltc.get_fc2_and_freqs(
                    ph3,
                    calculator=self.calculator,
                    pbar_kwargs={"leave": False, "disable": not prog_bar},
                )

                # Check for imaginary frequencies
                has_imaginary_freqs = check_imaginary_freqs(freqs)
                freqs_dict = {
                    Key.has_imag_ph_modes: has_imaginary_freqs,
                    Key.ph_freqs: freqs,
                }

                # If conductivity condition is met, calculate fc3
                ltc_condition = not has_imaginary_freqs and (
                    not relax_dict["broken_symmetry"] or conductivity_broken_symm
                )

                if ltc_condition:  # Calculate third-order force constants
                    logging.info(f"Calculating FC3 for {mat_id}")
                    ltc.calculate_fc3_set(
                        ph3,
                        calculator=self.calculator,
                        pbar_kwargs={"leave": False, "disable": not prog_bar},
                    )
                    ph3.produce_fc3(symmetrize_fc3r=True)

                if not ltc_condition:
                    results = info_dict | relax_dict | freqs_dict
                    warnings.warn(
                        f"{mat_id=} has imaginary frequencies or broken symmetry",
                        stacklevel=2,
                    )
                    all_results.append(results)
                    continue

            except Exception as exc:
                warnings.warn(
                    f"Failed to calculate force sets {mat_id}: {exc!r}", stacklevel=2
                )
                traceback.print_exc()
                info_dict["errors"].append(f"ForceConstantError: {exc!r}")
                info_dict["error_traceback"].append(traceback.format_exc())
                results = info_dict | relax_dict
                all_results.append(results)
                continue

            try:  # Calculate thermal conductivity
                logging.info(f"Calculating kappa for {mat_id}")
                ph3, kappa_dict, _cond = ltc.calculate_conductivity(
                    ph3, temperatures=temperatures
                )
            except Exception as exc:
                warnings.warn(
                    f"Failed to calculate conductivity {mat_id}: {exc!r}", stacklevel=2
                )
                traceback.print_exc()
                info_dict["errors"].append(f"ConductivityError: {exc!r}")
                info_dict["error_traceback"].append(traceback.format_exc())
                results = info_dict | relax_dict | freqs_dict
                all_results.append(results)
                continue

            results = info_dict | relax_dict | freqs_dict | kappa_dict
            all_results.append(results)

        return all_results

    def write_results(
        self,
        results: list[dict[str, Any]],
        results_dir: str,
        job_num: int = 0,
        num_jobs: int = 1,
    ) -> None:
        results_df = pd.DataFrame(results)
        results_df.to_json(
            os.path.join(
                results_dir,
                f"kappa103_dist{self.displacement}_{num_jobs}-{job_num}.json.gz",
            )
        )

    def save_state(self, checkpoint_location: str, is_preemption: bool = False) -> bool:
        return True



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

import logging

import torch


def init_edge_rot_euler_angles(edge_distance_vec):
    edge_vec_0 = edge_distance_vec
    edge_vec_0_distance = torch.sqrt(torch.sum(edge_vec_0**2, dim=1))

    # Make sure the atoms are far enough apart
    # assert torch.min(edge_vec_0_distance) < 0.0001
    if len(edge_vec_0_distance) > 0 and torch.min(edge_vec_0_distance) < 0.0001:
        logging.error(f"Error edge_vec_0_distance: {torch.min(edge_vec_0_distance)}")

    # make unit vectors
    xyz = edge_vec_0 / (edge_vec_0_distance.view(-1, 1))

    # are we standing at the north pole
    mask = xyz[:, 1].abs().isclose(xyz.new_ones(1))

    # compute alpha and beta

    # latitude (beta)
    beta = xyz.new_zeros(xyz.shape[0])
    beta[~mask] = torch.acos(xyz[~mask, 1])
    beta[mask] = torch.acos(xyz[mask, 1]).detach()

    # longitude (alpha)
    alpha = torch.zeros_like(beta)
    alpha[~mask] = torch.atan2(xyz[~mask, 0], xyz[~mask, 2])
    alpha[mask] = torch.atan2(xyz[mask, 0], xyz[mask, 2]).detach()

    # random gamma (roll)
    gamma = torch.rand_like(alpha) * 2 * torch.pi
    # gamma = torch.zeros_like(alpha)

    # intrinsic to extrinsic swap
    return -gamma, -beta, -alpha


# Borrowed from e3nn @ 0.4.0:
# https://github.com/e3nn/e3nn/blob/0.4.0/e3nn/o3/_wigner.py#L37
# In 0.5.0, e3nn shifted to torch.matrix_exp which is significantly slower:
# https://github.com/e3nn/e3nn/blob/0.5.0/e3nn/o3/_wigner.py#L92
def wigner_D(
    lv: int,
    alpha: torch.Tensor,
    beta: torch.Tensor,
    gamma: torch.Tensor,
    _Jd: list[torch.Tensor],
) -> torch.Tensor:
    alpha, beta, gamma = torch.broadcast_tensors(alpha, beta, gamma)
    J = _Jd[lv]
    Xa = _z_rot_mat(alpha, lv)
    Xb = _z_rot_mat(beta, lv)
    Xc = _z_rot_mat(gamma, lv)
    return Xa @ J @ Xb @ J @ Xc


def _z_rot_mat(angle: torch.Tensor, lv: int) -> torch.Tensor:
    M = angle.new_zeros((*angle.shape, 2 * lv + 1, 2 * lv + 1))

    # The following code needs to replaced for a for loop because
    # torch.export barfs on outer product like operations
    # ie: torch.outer(frequences, angle) (same as frequencies * angle[..., None])
    # will place a non-sense Guard on the dimensions of angle when attempting to export setting
    # angle (edge dimensions) as dynamic. This may be fixed in torch2.4.

    # inds = torch.arange(0, 2 * lv + 1, 1, device=device)
    # reversed_inds = torch.arange(2 * lv, -1, -1, device=device)
    # frequencies = torch.arange(lv, -lv - 1, -1, dtype=dtype, device=device)
    # M[..., inds, reversed_inds] = torch.sin(frequencies * angle[..., None])
    # M[..., inds, inds] = torch.cos(frequencies * angle[..., None])

    inds = list(range(0, 2 * lv + 1, 1))
    reversed_inds = list(range(2 * lv, -1, -1))
    frequencies = list(range(lv, -lv - 1, -1))
    for i in range(len(frequencies)):
        M[..., inds[i], reversed_inds[i]] = torch.sin(frequencies[i] * angle)
        M[..., inds[i], inds[i]] = torch.cos(frequencies[i] * angle)
    return M


def eulers_to_wigner(
    eulers: torch.Tensor,
    start_lmax: int,
    end_lmax: int,
    Jd: list[torch.Tensor],
) -> torch.Tensor:
    """
    set <rot_clip=True> to handle gradient instability when using gradient-based force/stress prediction.
    """
    alpha, beta, gamma = eulers

    size = int((end_lmax + 1) ** 2) - int((start_lmax) ** 2)
    wigner = torch.zeros(len(alpha), size, size, device=alpha.device, dtype=alpha.dtype)
    start = 0
    for lmax in range(start_lmax, end_lmax + 1):
        block = wigner_D(lmax, alpha, beta, gamma, Jd)
        end = start + block.size()[1]
        wigner[:, start:end, start:end] = block
        start = end

    return wigner



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.

ONLY for use in the NeurIPS 2021 Open Catalyst Challenge. For all other submissions
please use make_submission_file.py.
"""

from __future__ import annotations

import argparse
import glob
import os

import numpy as np


def write_is2re_relaxations(path: str, filename: str, hybrid) -> None:
    import ase.io
    from tqdm import tqdm

    submission_file = {}

    if not hybrid:
        ids = []
        energies = []
        systems = glob.glob(os.path.join(path, "*.traj"))
        for system in tqdm(systems):
            sid, _ = os.path.splitext(os.path.basename(system))
            ids.append(str(sid))
            traj = ase.io.read(system, "-1")
            energies.append(traj.get_potential_energy())

        submission_file["challenge_ids"] = np.array(ids)
        submission_file["challenge_energy"] = np.array(energies)

    else:
        preds = np.load(path)
        ids = []
        energies = []
        for sid, energy in zip(preds["ids"], preds["energy"]):
            ids.append(sid.split("_")[0])
            energies.append(energy)

        submission_file["challenge_ids"] = np.array(ids)
        submission_file["challenge_energy"] = np.array(energies)

    np.savez_compressed(filename, **submission_file)


def write_predictions(path: str, filename: str) -> None:
    submission_file = {}

    res = np.load(path, allow_pickle=True)
    contents = res.files
    for i in contents:
        key = f"challenge_{i}"
        submission_file[key] = res[i]

    np.savez_compressed(filename, **submission_file)


def main(args: argparse.Namespace) -> None:
    path = args.path

    if not args.out_path.endswith(".npz"):
        args.out_path = args.out_path + ".npz"

    if not args.is2re_relaxations:
        write_predictions(path, filename=args.out_path)
    else:
        write_is2re_relaxations(path, filename=args.out_path, hybrid=args.hybrid)
    print(f"Results saved to {args.out_path} successfully.")


if __name__ == "__main__":
    """
    Create a submission file for the NeurIPS 2021 Open Catalyst Challenge.

    Results file can be obtained as follows for the various tasks:

    S2EF: config["mode"] = "predict"
    IS2RE: config["mode"] = "predict"
    IS2RS: config["mode"] = "run-relaxations" and config["task"]["write_pos"] = True

    Use this script to write your results files in the format evalAI expects
    submissions.

    If writing IS2RE predictions from relaxations, the path specified must be a
    directory containg trajectory (.traj) files. Additionally, --is2re-relaxations must be
    provided as a command line argument.

    If writing IS2RE predictions from hybrid relaxations (force only model +
    energy only model), paths must be the .npz S2EF prediction files.
    Additionally, --is2re-relaxations and --hybrid must be provided as a
    command line argument.
    """

    parser = argparse.ArgumentParser()
    parser.add_argument("--path", help="Path to results")
    parser.add_argument("--out-path", help="Path to write predictions to.")
    parser.add_argument(
        "--is2re-relaxations",
        action="store_true",
        help="Write IS2RE results from trajectories. Path specified must be a directory containing .traj files.",
    )
    parser.add_argument(
        "--hybrid",
        action="store_true",
        help="Write IS2RE results from S2EF prediction files. Path specified must be a S2EF NPZ file.",
    )

    args: argparse.Namespace = parser.parse_args()
    main(args)



# client.py
from __future__ import annotations

import pickle

import websockets
from websockets.sync.client import connect as ws_connect


class AsyncMLIPInferenceWebSocketClient:
    def __init__(self, host, port, max_size=100 * 1024 * 1024):  # 100MB default
        self.uri = f"ws://{host}:{port}"
        self.websocket = None
        self.max_size = max_size

    async def connect(self):
        self.websocket = await websockets.connect(self.uri, max_size=self.max_size)

    async def close(self):
        await self.websocket.close()

    async def call(self, atomic_data):
        if not self.websocket:
            await self.connect()

        await self.websocket.send(pickle.dumps(atomic_data))

        # Receive response
        response = await self.websocket.recv()
        result = pickle.loads(response)
        return result


class SyncMLIPInferenceWebSocketClient:
    def __init__(self, host, port, max_size=100 * 1024 * 1024):  # 100MB default
        self.uri = f"ws://{host}:{port}"
        self.ws = None
        self.max_size = max_size

    def connect(self):
        self.ws = ws_connect(
            self.uri, ping_timeout=300, ping_interval=60, max_size=self.max_size
        )

    def call(self, atomic_data):
        if not self.ws:
            self.connect()

        self.ws.send(pickle.dumps(atomic_data))

        # Receive response
        response = self.ws.recv()
        result = pickle.loads(response)
        return result

    def __del__(self):
        self.close()

    def close(self):
        self.ws.close()



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

import inspect
import math
from bisect import bisect

import torch.optim.lr_scheduler as lr_scheduler


def warmup_lr_lambda(current_step: int, optim_config):
    """Returns a learning rate multiplier.
    Till `warmup_steps`, learning rate linearly increases to `initial_lr`,
    and then gets multiplied by `lr_gamma` every time a milestone is crossed.
    """

    # keep this block for older configs that have warmup_epochs instead of warmup_steps
    # and lr_milestones are defined in epochs
    if (
        any(x < 100 for x in optim_config["lr_milestones"])
        or "warmup_epochs" in optim_config
    ):
        raise Exception(
            "ConfigError: please define lr_milestones in steps not epochs and define warmup_steps instead of warmup_epochs"
        )

    if current_step <= optim_config["warmup_steps"]:
        alpha = current_step / float(optim_config["warmup_steps"])
        return optim_config["warmup_factor"] * (1.0 - alpha) + alpha
    else:
        idx = bisect(optim_config["lr_milestones"], current_step)
        return pow(optim_config["lr_gamma"], idx)


class CosineLRLambda:
    def __init__(
        self,
        warmup_epochs: int,
        warmup_factor: float,
        epochs: int,
        lr_min_factor: float,
    ) -> None:
        self.warmup_epochs = warmup_epochs
        self.lr_warmup_factor = warmup_factor
        self.max_epochs = epochs
        self.lr_min_factor = lr_min_factor

    def __call__(self, current_step: int) -> float:
        # `warmup_epochs` is already multiplied with the num of iterations
        if current_step <= self.warmup_epochs:
            alpha = current_step / float(self.warmup_epochs)
            return self.lr_warmup_factor * (1.0 - alpha) + alpha
        else:
            if current_step >= self.max_epochs:
                return self.lr_min_factor
            return self.lr_min_factor + 0.5 * (1 - self.lr_min_factor) * (
                1 + math.cos(math.pi * (current_step / self.max_epochs))
            )


class LRScheduler:
    """
    Learning rate scheduler class for torch.optim learning rate schedulers

    Notes:
        If no learning rate scheduler is specified in the config the default
        scheduler is warmup_lr_lambda (fairchem.core.common.utils) not no scheduler,
        this is for backward-compatibility reasons. To run without a lr scheduler
        specify scheduler: "Null" in the optim section of the config.

    Args:
        optimizer (obj): torch optim object
        config (dict): Optim dict from the input config
    """

    def __init__(self, optimizer, config) -> None:
        self.optimizer = optimizer
        self.config = config.copy()
        if "scheduler" in self.config:
            self.scheduler_type = self.config["scheduler"]
        else:
            self.scheduler_type = "LambdaLR"

            def scheduler_lambda_fn(x):
                return warmup_lr_lambda(x, self.config)

            self.config["lr_lambda"] = scheduler_lambda_fn
        if self.scheduler_type != "Null":
            self.scheduler = getattr(lr_scheduler, self.scheduler_type)
            scheduler_args = self.filter_kwargs(config)
            self.scheduler = self.scheduler(optimizer, **scheduler_args)

    def step(self, metrics=None, epoch=None) -> None:
        if self.scheduler_type == "Null":
            return
        if self.scheduler_type == "ReduceLROnPlateau":
            if metrics is None:
                raise Exception("Validation set required for ReduceLROnPlateau.")
            self.scheduler.step(metrics)
        else:
            self.scheduler.step()

    def filter_kwargs(self, config):
        # adapted from https://stackoverflow.com/questions/26515595/
        sig = inspect.signature(self.scheduler)
        filter_keys = [
            param.name
            for param in sig.parameters.values()
            if param.kind == param.POSITIONAL_OR_KEYWORD
        ]
        filter_keys.remove("optimizer")
        return {arg: self.config[arg] for arg in self.config if arg in filter_keys}

    def get_lr(self):
        for group in self.optimizer.param_groups:
            return group["lr"]
        return None



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.

Uncompresses downloaded S2EF datasets to be used by the LMDB preprocessing
script - preprocess_ef.py
"""

from __future__ import annotations

import argparse
import glob
import lzma
import multiprocessing as mp
import os

from tqdm import tqdm


def read_lzma(inpfile: str, outfile: str) -> None:
    with open(inpfile, "rb") as f:
        contents = lzma.decompress(f.read())
        with open(outfile, "wb") as op:
            op.write(contents)


def decompress_list_of_files(ip_op_pair: tuple[str, str]) -> None:
    ip_file, op_file = ip_op_pair
    read_lzma(ip_file, op_file)


def get_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--ipdir", type=str, help="Path to compressed dataset directory"
    )
    parser.add_argument(
        "--opdir", type=str, help="Directory path to uncompress files to"
    )
    parser.add_argument(
        "--num-workers", type=int, help="# of processes to parallelize across"
    )
    return parser


def main(args: argparse.Namespace) -> None:
    os.makedirs(args.opdir, exist_ok=True)

    filelist = glob.glob(os.path.join(args.ipdir, "*txt.xz")) + glob.glob(
        os.path.join(args.ipdir, "*extxyz.xz")
    )
    ip_op_pairs: list[tuple[str, str]] = []
    for filename in filelist:
        fname_base = os.path.basename(filename)
        ip_op_pairs.append((filename, os.path.join(args.opdir, fname_base[:-3])))

    pool = mp.Pool(args.num_workers)
    list(
        tqdm(
            pool.imap(decompress_list_of_files, ip_op_pairs),
            total=len(ip_op_pairs),
            desc=f"Uncompressing {args.ipdir}",
        )
    )


if __name__ == "__main__":
    parser: argparse.ArgumentParser = get_parser()
    args: argparse.Namespace = parser.parse_args()
    main(args)



"""
Sequential request server with parallel model execution
Usage: python server.py --workers 4 --port 8000
"""

from __future__ import annotations

import asyncio
import contextlib
import logging
import pickle
import signal
from typing import TYPE_CHECKING

import hydra
import torch.distributed as dist
import websockets
from monty.dev import requires
from torch.distributed.elastic.utils.distributed import get_free_port
from websockets.asyncio.server import serve

if TYPE_CHECKING:
    from omegaconf import DictConfig

from fairchem.core.common import gp_utils
from fairchem.core.common.distutils import (
    assign_device_for_local_rank,
    get_device_for_local_rank,
    setup_env_local_multi_gpu,
)

try:
    import ray
    from ray import remote

    ray_installed = True
except ImportError:
    ray = None

    def remote(cls):
        # dummy
        return cls

    ray_installed = False

logging.basicConfig(level=logging.INFO)


@remote
class MLIPWorker:
    def __init__(
        self, worker_id: int, world_size: int, master_port: int, predictor_config: dict
    ):
        if ray_installed is False:
            raise RuntimeError("Requires `ray` to be installed")

        self.worker_id = worker_id
        self._distributed_setup(
            worker_id, master_port, world_size, predictor_config.get("device", "cpu")
        )
        self.predict_unit = hydra.utils.instantiate(predictor_config)
        logging.info(
            f"Worker {worker_id}, gpu_id: {ray.get_gpu_ids()}, loaded predict unit: {self.predict_unit}, "
            f"on port {master_port}, with device: {get_device_for_local_rank()}, config: {predictor_config}"
        )

    def _distributed_setup(
        self, worker_id: int, master_port: int, world_size: int, device: str
    ):
        # initialize distributed environment
        # TODO, this wont work for multi-node, need to fix master addr
        setup_env_local_multi_gpu(worker_id, master_port)
        # local_rank = int(os.environ["LOCAL_RANK"])
        assign_device_for_local_rank(device == "cpu", 0)
        backend = "gloo" if device == "cpu" else "nccl"
        dist.init_process_group(
            backend=backend,
            rank=worker_id,
            world_size=world_size,
        )
        gp_utils.setup_graph_parallel_groups(world_size, backend)

    def predict(self, data: bytes):
        atomic_data = pickle.loads(data)
        result = self.predict_unit.predict(atomic_data)
        return pickle.dumps(result)


@requires(ray_installed, message="Requires `ray` to be installed")
class MLIPInferenceServerWebSocket:
    def __init__(self, predictor_config: dict, port=8001, num_workers=1):
        logging.basicConfig(level=logging.INFO)
        self.host = "localhost"
        self.port = port
        self.num_workers = num_workers
        self.predictor_config = predictor_config
        # Initialize a pool of MLIPWorkers
        self.master_pg_port = get_free_port()
        ray.init(logging_level=logging.INFO)
        options = {"num_gpus": 1} if predictor_config.get("device") == "cuda" else {}
        self.workers = [
            MLIPWorker.options(**options).remote(
                i, self.num_workers, self.master_pg_port, self.predictor_config
            )
            for i in range(self.num_workers)
        ]
        logging.info(
            "Initialized Local MLIPInferenceServerWebSocket with config: "
            f"{self.predictor_config}, port: {self.port}, workers: {self.num_workers}"
        )

        # Set up signal handlers for clean shutdown
        self._setup_signal_handlers()

    def _setup_signal_handlers(self):
        """Set up signal handlers for graceful shutdown"""

        def signal_handler(signum, frame):
            logging.info(f"Received signal {signum}, initiating shutdown...")
            self.shutdown()

        signal.signal(signal.SIGTERM, signal_handler)
        signal.signal(signal.SIGINT, signal_handler)

    async def handler(self, websocket):
        try:
            async for message in websocket:
                # don't unpickle here, just pass bytes to workers
                futures = [w.predict.remote(message) for w in self.workers]
                # just get the first result that is ready since they are identical
                # the rest of the futures should go out of scope and memory garbage collected
                ready_ids, _ = ray.wait(futures, num_returns=1)
                await websocket.send(ray.get(ready_ids[0]))
        except websockets.exceptions.ConnectionClosed:
            logging.info("Client disconnected")
        except Exception as e:
            logging.info(f"MLIPInferenceServer handler Error: {e}")
        finally:
            self.shutdown()

    async def start(self):
        self.stop_event = asyncio.Event()

        async with serve(self.handler, self.host, self.port, max_size=10 * 1024 * 1024):
            print(f"WebSocket server started on port {self.port}")

            with contextlib.suppress(asyncio.CancelledError):
                await self.stop_event.wait()

    def run(self):
        """Run the server (blocking)"""
        asyncio.run(self.start())

    def shutdown(self):
        """Shutdown the server and clean up Ray resources"""
        if hasattr(self, "stop_event"):
            self.stop_event.set()
        ray.shutdown()
        logging.info("MLIPInferenceServerWebSocket shutdown complete")


@hydra.main(
    version_base=None,
    config_path=".",
    config_name="server_config",
)
def main(cfg: DictConfig):
    # Ensure logs from all Ray workers are printed to the driver
    server = MLIPInferenceServerWebSocket(
        cfg.predict_unit, cfg.server.port, cfg.server.workers
    )
    server.run()


if __name__ == "__main__":
    main()



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

import torch
import torch.nn as nn
import torch.nn.functional as F


class ScaledSiLU(nn.Module):
    def __init__(self, inplace: bool = False) -> None:
        super().__init__()
        self.inplace = inplace
        self.scale_factor = 1.6791767923989418

    def forward(self, inputs):
        return F.silu(inputs, inplace=self.inplace) * self.scale_factor

    def extra_repr(self):
        str = f"scale_factor={self.scale_factor}"
        if self.inplace:
            str = str + ", inplace=True"
        return str


# Reference: https://github.com/facebookresearch/llama/blob/main/llama/model.py#L175
class ScaledSwiGLU(nn.Module):
    def __init__(self, in_channels: int, out_channels: int, bias: bool = True) -> None:
        super().__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.w = torch.nn.Linear(in_channels, 2 * out_channels, bias=bias)
        self.act = ScaledSiLU()

    def forward(self, inputs):
        w = self.w(inputs)
        w_1 = w.narrow(-1, 0, self.out_channels)
        w_1 = self.act(w_1)
        w_2 = w.narrow(-1, self.out_channels, self.out_channels)
        return w_1 * w_2


# Reference: https://github.com/facebookresearch/llama/blob/main/llama/model.py#L175
class SwiGLU(nn.Module):
    def __init__(self, in_channels: int, out_channels: int, bias: bool = True) -> None:
        super().__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.w = torch.nn.Linear(in_channels, 2 * out_channels, bias=bias)
        self.act = torch.nn.SiLU()

    def forward(self, inputs):
        w = self.w(inputs)
        w_1 = w.narrow(-1, 0, self.out_channels)
        w_1 = self.act(w_1)
        w_2 = w.narrow(-1, self.out_channels, self.out_channels)
        return w_1 * w_2


class SmoothLeakyReLU(torch.nn.Module):
    def __init__(self, negative_slope: float = 0.2) -> None:
        super().__init__()
        self.alpha = negative_slope

    def forward(self, x):
        x1 = ((1 + self.alpha) / 2) * x
        x2 = ((1 - self.alpha) / 2) * x * (2 * torch.sigmoid(x) - 1)
        return x1 + x2

    def extra_repr(self):
        return f"negative_slope={self.alpha}"


class ScaledSmoothLeakyReLU(torch.nn.Module):
    def __init__(self) -> None:
        super().__init__()
        self.act = SmoothLeakyReLU(0.2)
        self.scale_factor = 1.531320475574866

    def forward(self, x):
        return self.act(x) * self.scale_factor

    def extra_repr(self):
        return f"negative_slope={self.act.alpha}, scale_factor={self.scale_factor}"


class ScaledSigmoid(torch.nn.Module):
    def __init__(self) -> None:
        super().__init__()
        self.scale_factor = 1.8467055342154763

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return torch.sigmoid(x) * self.scale_factor


class GateActivation(torch.nn.Module):
    # m_prime -> order is l0m0, l1m0, l2m0.. , l1m1 , l2m1, ... , l1m-1, l2m-1,...
    def __init__(
        self, lmax: int, mmax: int, num_channels: int, m_prime: bool = False
    ) -> None:
        super().__init__()
        self.lmax = lmax
        self.mmax = mmax
        self.num_channels = num_channels

        # compute `expand_index` based on `lmax` and `mmax`
        num_components = 0
        for lval in range(1, self.lmax + 1):
            num_m_components = min((2 * lval + 1), (2 * self.mmax + 1))
            num_components = num_components + num_m_components
        expand_index = torch.zeros([num_components]).long()

        self.m_prime = m_prime
        if self.m_prime:
            start_idx = 0
            length = self.lmax
            expand_index[start_idx : (start_idx + length)] = torch.arange(self.lmax)
            start_idx = start_idx + length
            for mval in range(1, self.mmax + 1):
                length = 2 * (self.lmax + 1 - mval)
                expand_index[start_idx : (start_idx + length)] = torch.cat(
                    [
                        torch.arange(mval - 1, self.lmax),
                        torch.arange(mval - 1, self.lmax),
                    ]
                )
                start_idx = start_idx + length
        else:
            start_idx = 0
            for lval in range(1, self.lmax + 1):
                length = min((2 * lval + 1), (2 * self.mmax + 1))
                expand_index[start_idx : (start_idx + length)] = lval - 1
                start_idx = start_idx + length
        self.register_buffer("expand_index", expand_index, persistent=False)

        self.scalar_act = (
            torch.nn.SiLU()
        )  # SwiGLU(self.num_channels, self.num_channels)  # #
        self.gate_act = torch.nn.Sigmoid()  # torch.nn.SiLU() # #

    def forward(self, gating_scalars, input_tensors):
        """
        `gating_scalars`: shape [N, lmax * num_channels]
        `input_tensors`: shape  [N, (lmax + 1) ** 2, num_channels]
        """
        gating_scalars = self.gate_act(gating_scalars).view(
            gating_scalars.shape[0], self.lmax, self.num_channels
        )

        gating_scalars = torch.index_select(
            gating_scalars, dim=1, index=self.expand_index
        )
        input_tensors_scalars, input_tensors_vectors = input_tensors.split(
            (1, input_tensors.shape[1] - 1), 1
        )

        input_tensors_scalars = self.scalar_act(input_tensors_scalars)
        input_tensors_vectors = input_tensors_vectors * gating_scalars

        return torch.cat((input_tensors_scalars, input_tensors_vectors), dim=1)


class S2Activation(torch.nn.Module):
    """
    Assume we only have one resolution
    """

    def __init__(self, lmax: int, mmax: int, SO3_grid) -> None:
        super().__init__()
        self.lmax = lmax
        self.mmax = mmax
        self.act = torch.nn.SiLU()
        self.SO3_grid = SO3_grid

    def forward(self, inputs):
        to_grid_mat = self.SO3_grid["lmax_mmax"].get_to_grid_mat()
        from_grid_mat = self.SO3_grid["lmax_mmax"].get_from_grid_mat()
        x_grid = torch.einsum("bai, zic -> zbac", to_grid_mat, inputs)
        x_grid = self.act(x_grid)
        return torch.einsum("bai, zbac -> zic", from_grid_mat, x_grid)


class SeparableS2Activation(torch.nn.Module):
    def __init__(self, lmax: int, mmax: int, SO3_grid) -> None:
        super().__init__()
        self.lmax = lmax
        self.mmax = mmax
        self.scalar_act = torch.nn.SiLU()
        self.s2_act = S2Activation(self.lmax, self.mmax, SO3_grid)

    def forward(self, input_scalars, input_tensors):
        output_scalars = self.scalar_act(input_scalars)
        output_scalars = output_scalars.reshape(
            output_scalars.shape[0], 1, output_scalars.shape[-1]
        )
        output_tensors = self.s2_act(input_tensors)
        return torch.cat(
            (
                output_scalars,
                output_tensors.narrow(1, 1, output_tensors.shape[1] - 1),
            ),
            dim=1,
        )


class S2Activation_M(torch.nn.Module):
    """
    Assume we only have one resolution
    """

    def __init__(self, lmax: int, mmax: int, SO3_grid, to_m: torch.Tensor) -> None:
        super().__init__()
        self.lmax = lmax
        self.mmax = mmax
        self.act = torch.nn.SiLU()
        self.SO3_grid = SO3_grid
        to_grid_mat = self.SO3_grid["lmax_mmax"].get_to_grid_mat()
        to_grid_mat_m = torch.einsum("ji,bai->jba", to_m, to_grid_mat)
        self.register_buffer("to_grid_mat_m", to_grid_mat_m, persistent=False)
        from_grid_mat = self.SO3_grid["lmax_mmax"].get_from_grid_mat()
        from_grid_mat_m = torch.einsum("ji,bai->baj", to_m, from_grid_mat)
        self.register_buffer("from_grid_mat_m", from_grid_mat_m, persistent=False)

    def forward(self, inputs):
        x_grid = torch.einsum("iba, zic -> zbac", self.to_grid_mat_m, inputs)
        x_grid = self.act(x_grid)
        return torch.einsum("bai, zbac -> zic", self.from_grid_mat_m, x_grid)


class SeparableS2Activation_M(torch.nn.Module):
    def __init__(self, lmax: int, mmax: int, SO3_grid, to_m) -> None:
        super().__init__()
        self.lmax = lmax
        self.mmax = mmax
        self.scalar_act = torch.nn.SiLU()
        self.s2_act = S2Activation_M(self.lmax, self.mmax, SO3_grid, to_m)

    def forward(self, input_scalars, input_tensors):
        output_scalars = self.scalar_act(input_scalars)
        output_scalars = output_scalars.reshape(
            output_scalars.shape[0], 1, output_scalars.shape[-1]
        )
        output_tensors = self.s2_act(input_tensors)
        return torch.cat(
            (
                output_scalars,
                output_tensors.narrow(1, 1, output_tensors.shape[1] - 1),
            ),
            dim=1,
        )



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

import os
from glob import glob

import numpy as np
import pandas as pd

from fairchem.core.components.benchmark.benchmark_reducer import BenchmarkReducer
from fairchem.core.components.calculate.nve_md_runner import NVEMDRunner, get_thermo


def moving_avg(x, window=20):
    all_x = []
    for i in range(window):
        all_x.append(x[i : len(x) - (window - i)])  # noqa: PERF401
    return np.stack(all_x).mean(axis=0)


def get_te_drift(filename):
    time, Et = get_thermo(filename)
    total_step = len(time)
    # remove 10% for equilibration
    equil_time = int(total_step * 0.1)
    Et = Et[equil_time:]
    total_time = time[-1] - time[equil_time]
    Et = moving_avg(Et, window=20)
    drift = 1000 * np.abs(Et[-1] - Et[equil_time]) / total_time
    return float(drift)


class NVEMDReducer(BenchmarkReducer):
    def __init__(
        self,
        benchmark_name: str,
    ):
        """
        Args:
            benchmark_name: Name of the benchmark, used for file naming
            target_data_path: Path to the target data JSON file
            index_name: Optional name of the column to use as index
            corrections: Optional correction class to apply to all entries
            max_error_threshold: Maximum allowed mean absolute formation energy per atom error threshold
        """
        self.benchmark_name = benchmark_name

    @property
    def runner_type(self) -> type[NVEMDRunner]:
        """The runner type this reducer is associated with."""
        return NVEMDRunner

    def join_results(self, results_dir: str, glob_pattern: str) -> pd.DataFrame:
        """Join results from multiple JSON files into a single DataFrame.

        Args:
            results_dir: Directory containing result files
            glob_pattern: Pattern to match result files

        Returns:
            Combined DataFrame containing all results
        """
        results = [
            get_te_drift(f) for f in glob(os.path.join(results_dir, glob_pattern))
        ]
        return results

    def save_results(self, results: list, results_dir: str) -> None:
        """Save joined results to a compressed json file

        Args:
            results:  results: Combined results from join_results
            results_dir: Directory containing result files
        """
        with open(
            os.path.join(results_dir, f"{self.benchmark_name}_results.txt"), "w"
        ) as f:
            f.write("\n".join([str(x) for x in results]))

    def compute_metrics(self, results: list, run_name: str) -> pd.DataFrame:
        """Compute Matbench discovery metrics for relaxed energy and structure predictions.

        Args:
            results: DataFrame containing prediction results with energy values
            run_name: Identifier for the current evaluation run

        Returns:
            DataFrame containing computed metrics for different material subsets
        """
        metrics = {
            "drift_mev_per_atom_ps": float(np.mean(results)),
        }
        return pd.DataFrame([metrics], index=[run_name])

    def save_metrics(self, metrics: pd.DataFrame, results_dir: str) -> None:
        """Save computed metrics to a compressed JSON file.

        Args:
            metrics: DataFrame containing the computed metrics
            results_dir: Directory where metrics will be saved
        """
        metrics.to_json(
            os.path.join(results_dir, f"{self.benchmark_name}_metrics.json")
        )

    def log_metrics(self, metrics: pd.DataFrame, run_name: str) -> None:
        """Log metrics to the configured logger if available.

        Args:
            metrics: DataFrame containing the computed metrics
            run_name: Name of the current run
        """
        if self.logger is not None:
            self.logger.log_dataframe(
                name=self.benchmark_name, dataframe=metrics.reset_index()
            )

    def save_state(self, checkpoint_location: str, is_preemption: bool = False) -> bool:
        pass

    def load_state(self, checkpoint_location: str | None) -> None:
        pass
